from ase.db import connect
from typing import List, Dict, Set


def get_all_cf_names(db_name: str, tab_name: str) -> Set[str]:
    """
    Return a list with all correlation function names

    Parameters:

    db_name: Name of the database

    tab_name: Name of the table in the database where correlation functions
        are stores
    """
    with connect(db_name) as db:
        query = f"SELECT key FROM {tab_name}"
        cur = db.connection.cursor()
        db_names = set()
        cur.execute(query)
        for row in cur.fetchall():
            db_names.add(row[0])
    return db_names


def get_all_cf(db_name: str, tab_name: str, db_id: int) -> Dict[str, float]:
    """
    Return all correlation functions associated with an entry in the database

    Parameters:

    db_name: Name of the database

    tab_name: Name of the table where the correlation functions are stored

    db_id: ID of the entry
    """
    with connect(db_name) as db:
        query = f"SELECT key, value FROM {tab_name} WHERE id=?"
        cur = db.connection.cursor()
        cur.execute(query, (db_id,))
        cf = {}
        for row in cur.fetchall():
            cf_name, cf_value = row
            cf[cf_name] = cf_value
    return cf


def get_cf_tables(db_name: str) -> List[str]:
    """
    Return a list with table names that contain correlation functions

    Parameter:

    db_name: Name of the database
    """
    db = connect(db_name)
    ext_tab = db._get_external_table_names()
    cf_tables = [n for n in ext_tab if n.endswith('_cf')]
    return cf_tables



# pylint: disable=too-many-lines
"""A collection of miscellaneous functions used for Cluster Expansion."""
import re
import logging
from pathlib import Path
from itertools import (permutations, combinations, product, filterfalse, chain)
from collections.abc import Iterable
from typing import List, Optional, Tuple, Dict, Set, Sequence, NamedTuple, Union
from typing import Iterable as tIterable
from typing_extensions import Protocol
import numpy as np
from numpy.random import sample, shuffle
from ase.db import connect
from ase.db.core import parse_selection
from scipy.spatial import cKDTree as KDTree
from scipy.optimize import linprog

logger = logging.getLogger(__name__)


# pylint: disable=too-few-public-methods
class ApproxEqualityList:
    """
    Wrapper around a list which implements a new comparison operator. If two
    items in the list is equal within a given tolerance, the items are
    considered to be equal.

    Parameters:

    array: list
        List that is wrapped

    tol: float
        Toleracnce for comparison check
    """

    def __init__(self, array, tol=1E-5):
        self.array = array
        self.tol = tol

    def __lt__(self, other):
        for x, y in zip(self.array, other.array):
            if x - y < -self.tol:
                return True
            if x - y > self.tol:
                return False
        return False


def index_by_position(atoms):
    """Set atomic indices by its position."""
    # add zero to avoid negative zeros
    tags = atoms.get_positions() + 0
    tags = tags.tolist()
    tags = [ApproxEqualityList(x) for x in tags]
    deco = sorted([(tag, i) for i, tag in enumerate(tags)])

    indices = [i for tag, i in deco]
    return indices


def sort_by_position(atoms):
    """Sort atoms by its position."""
    # Return a new Atoms object with sorted atomic order.
    # The default is to order according to chemical symbols,
    # but if *tags* is not None, it will be used instead.
    # A stable sorting algorithm is used.
    indices = index_by_position(atoms)
    return atoms[indices]


def wrap_and_sort_by_position(atoms):
    """Wrap and sort atoms by their positions."""
    atoms.wrap()
    atoms = sort_by_position(atoms)
    return atoms


def ndarray2list(data):
    """
    Convert nested lists of a combination of lists and numpy arrays
    to list of lists
    """
    if not isinstance(data, list) and not isinstance(data, np.ndarray):
        return data

    data = list(data)
    for i, value in enumerate(data):
        data[i] = ndarray2list(value)
    return list(data)


def dec_string(deco, equiv_sites):
    """Create the decoration string based on equiv sites."""
    equiv_dec = sorted(equivalent_deco(deco, equiv_sites))
    return ''.join(str(i) for i in equiv_dec[0])


def equivalent_deco(deco, equiv_sites):
    """Generate equivalent decoration numbers based on equivalent sites."""
    if not equiv_sites:
        return [deco]

    perm = []
    for equiv in equiv_sites:
        perm.append(list(permutations(equiv)))

    equiv_deco = []
    for comb in product(*perm):
        order = []
        for item in comb:
            order += list(item)

        orig_order = list(range(len(deco)))
        for i, srt_indx in enumerate(sorted(order)):
            orig_order[srt_indx] = order[i]
        equiv_deco.append([deco[indx] for indx in orig_order])

    unique_deco = []
    for eq_dec in equiv_deco:
        if eq_dec not in unique_deco:
            unique_deco.append(eq_dec)
    return unique_deco


def flatten(x):
    """Flatten list."""
    if isinstance(x, Iterable):
        return [a for i in x for a in flatten(i)]
    return [x]


def nested_array2list(array):
    """Convert a nested array/tuple to a nested list."""
    if isinstance(array, np.ndarray):
        array = array.tolist()
    else:
        array = list(array)
    try:
        for i, value in enumerate(array):
            if isinstance(value, np.ndarray):
                array[i] = value.tolist()
            else:
                array[i] = list(value)
    except TypeError:
        pass
    return array


def update_db(uid_initial=None,
              final_struct=None,
              db_name=None,
              custom_kvp_init: dict = None,
              custom_kvp_final: dict = None):
    """Update the database.

    Parameters:

    uid_initial: int
        entry ID of the initial structure in the database

    final_struct: Atoms
        Atoms object with the final structure with a physical
        quantity that needs to be modeled (e.g., DFT energy)

    db_name: str
        Database name

    custom_kvp_init: dict (optional)
        If desired, one can pass additional key-value-pairs for the
        entry containing the initial structure

    custom_kvp_final: dict (optional)
        If desired, one can pass additional key-value-pairs for the
        entry containing the final structure
    """
    if custom_kvp_init is None:
        custom_kvp_init = {}
    if custom_kvp_final is None:
        custom_kvp_final = {}

    db = connect(db_name)

    init_row = db.get(id=uid_initial)

    # Check if a final structure already exits
    name = init_row.name
    select_cond = [('name', '=', name), ('struct_type', '=', 'final')]
    exist = sum(1 for row in db.select(select_cond))
    if exist >= 1:
        logger.warning("A structure with 'name'=%s and 'struct_type'=final already exits in DB.",
                       name)
        return

    # Write the final structure to database
    kvp_final = {'struct_type': 'final', 'name': name}
    kvp_final.update(custom_kvp_final)
    uid_final = db.write(final_struct, key_value_pairs=kvp_final)

    kvp_update_init = {
        'converged': True,
        'started': '',
        'queued': '',
    }
    if kvp_final['struct_type'] == 'final':
        kvp_update_init['final_struct_id'] = uid_final
    kvp_update_init.update(custom_kvp_init)

    # Update info for the initial structure
    db.update(uid_initial, **kvp_update_init)


def exclude_information_entries():
    """Return selection condition to exlcude all entries in the database that
       only contain information about the clusters.
    """
    return [('name', '!=', 'primitive_cell'), ('name', '!=', 'template')]


def get_all_internal_distances(atoms, max_dist, ref_indices):
    """Obtain all internal distances of the passed atoms object and return a
       Numpy array containing all the distances sorted in an ascending order.
    """
    tree = KDTree(atoms.get_positions())
    distances = []
    for ind in ref_indices:
        indices = tree.query_ball_point(atoms[ind].position, max_dist)
        dists = atoms.get_distances(ind, indices)
        for d in dists:
            if np.any(np.abs(np.array(distances) - d) < 1E-6):
                continue
            distances.append(d)
    distances = sorted(distances)

    # Make sure that the first element is 0
    assert distances[0] < 1E-6
    return np.array(distances[1:])


def reconfigure(settings, select_cond=None):
    from clease.corr_func import CorrFunction  # pylint: disable=import-outside-toplevel
    CorrFunction(settings).reconfigure_db_entries(select_cond)


def split_dataset(X: np.ndarray,
                  y: np.ndarray,
                  nsplits: int = 10,
                  groups: Sequence[int] = ()) -> List[Dict[str, np.ndarray]]:
    """Split the dataset such that it can be used for k-fold
        cross validation.

    :param X: Design matrix
    :param y: Target values
    :param nsplits: Number of partittions
    :param groups: List of the same length as y. Each row
        in y with the same group tag, is treated as the same
        group. Datapoints in the same group are never split
        across different partitions. If an empty list is give,
        each item in y is assumed to constitute its own group.
    """
    if not groups:
        groups = list(range(len(y)))
    unique_groups = list(set(groups))

    if len(unique_groups) < nsplits:
        raise ValueError("The number of unique groups has to be greater "
                         "than the number of partitions.")
    shuffle(unique_groups)
    partitions = []
    num_validation = int(len(unique_groups) / nsplits)

    if num_validation < 1:
        num_validation = 1
    for i in range(nsplits):
        start = i * num_validation
        end = (i + 1) * num_validation
        if i == nsplits - 1:
            chosen_groups = unique_groups[start:]
        else:
            chosen_groups = unique_groups[start:end]

        group_mask = np.zeros(len(groups), dtype=np.uint8)
        group_mask[chosen_groups] = 1
        index_mask = np.zeros(len(y), dtype=np.uint8)
        for j, g in enumerate(groups):
            if group_mask[g]:
                index_mask[j] = 1
        data = {
            "train_X": X[index_mask == 0, :],
            "train_y": y[index_mask == 0],
            "validate_X": X[index_mask == 1, :],
            "validate_y": y[index_mask == 1]
        }
        partitions.append(data)
    return partitions


def random_validation_set(num: int = 10,
                          select_cond: Optional[list] = None,
                          db_name: Optional[str] = None):
    """
    Construct a random test set.

    :param num: Number of datapoints to include in the test set
    :param select_cond: Select condition to be used to select items from
        the database. If not given, it will be struct_type='initial',
        converged=True
    :param db_name: Name of the database
    """
    if select_cond is None:
        select_cond = [('struct_type', '=', 'initial'), ('converged', '=', True)]

    if db_name is None:
        raise ValueError("No database provided!")

    all_ids = get_ids(select_cond, db_name)
    return sample(all_ids, num)


def exclude_ids(ids: List[int]) -> List[tuple]:
    """
    Construct a select condition based on the ids passed.

    :params ids: List of IDs
    """
    return [("id", "!=", x) for x in ids]


def symbols2integer(basis_functions):
    """Convert each symbol in the basis function
       to a unique integer.

    Parameters:

    basis_functions: list of dict
        The basis function dictionary from ClusterExpansionSettings
    """
    symb_id = {}
    for i, symb in enumerate(basis_functions[0].keys()):
        symb_id[symb] = i
    return symb_id


def bf2npyarray(basis_functions, symb_id):
    """Convert the basis function dictionary to a 2D
        numpy array

    Parameters:

    basis_function: list of dict
        Basis function dictionary from ClusterExpansionSettings

    symb_id: dict
        Dictionary of symbol integer (see symbols2integer)
    """
    bf_npy = np.zeros((len(basis_functions), len(symb_id)))

    for i, bf in enumerate(basis_functions):
        for k, v in bf.items():
            bf_npy[i, symb_id[k]] = v
    return bf_npy


def nested_list2str(nested_list):
    """Convert a nested list to string."""
    return 'x'.join(','.join(str(x) for x in item) for item in nested_list)


def str2nested_list(string):
    """Convert string to nested list."""

    def _as_int(x):
        return int(x)

    return [list(map(_as_int, item.split(','))) for item in string.split('x')]


def min_distance_from_facet(x, cell):
    """
    Calculate the minimum distance from a point to the cell facet.

    Parameters:

    x: np.array
        Position from which to calculate the minimum distance

    cell: Cell
        Cell of an Atoms object
    """
    dists = []

    for plane in combinations([0, 1, 2], r=2):
        # Unit normal vector
        n = np.cross(cell[plane[0], :], cell[plane[1], :])
        n /= np.sqrt(n.dot(n))

        # Plane with origin in it
        dist = np.abs(n.dot(x))
        dists.append(dist)

        # Opposite facet
        remaining = list(set([0, 1, 2]) - set(plane))[0]
        vec = cell[remaining, :]
        dist = np.abs(n.dot(x - vec))
        dists.append(dist)
    return min(dists)


def trans_matrix_index2tags(trans_matrix, tagged_atoms, indices=None):
    """
    Convert from indices to tags

    Parameters:

    trans_matrix: list of dict
        Original translation matrix

    tagged_atoms: Atoms
        Atoms with a tag that should be used instead of the
        index

    indices: list of int
        Atom indices corresponding to each row in trans_matrix.
        If None, it is assumed that len(trans_matrix) == len(tagged_atoms)
        and each row in trans_matrix corresponds to the atom with the same
        index in tagged_atoms.
    """
    unique_tags = sorted(list(set(atom.tag for atom in tagged_atoms)))

    if indices is None:
        indices = list(range(len(trans_matrix)))

    # Make sure we have a continuous series of tags
    assert len(unique_tags) == max(unique_tags) + 1

    new_tm = [{} for _ in range(len(unique_tags))]
    used_tags = [False for _ in range(len(unique_tags))]

    for i, row in enumerate(trans_matrix):
        tag = tagged_atoms[indices[i]].tag
        if used_tags[tag]:
            continue
        new_row = {tagged_atoms[k].tag: tagged_atoms[v].tag for k, v in row.items()}
        used_tags[tag] = True
        new_tm[tag] = new_row
    return new_tm


def indices2tags(supercell, clusters):
    for cluster in clusters:
        for i, figure in enumerate(cluster):
            cluster[i] = [int(supercell[x].tag) for x in figure]
    return clusters


def list2str(array):
    return "-".join(str(x) for x in array)


def factorize(n):
    while n > 1:
        for i in range(2, n + 1):
            if n % i == 0:
                n = int(n / i)
                yield i
                break


def count_atoms(atoms):
    """
    Return dictionary with the number of items of each species
    """
    count = {}
    for s in atoms.symbols:
        count[s] = count.get(s, 0) + 1
    return count


def all_integer_transform_matrices_given_diag(diag):
    rng1 = range(0, diag[0] + 1)
    rng2 = rng1
    rng3 = range(0, diag[1] + 1)
    for off_diag in product(rng1, rng2, rng3):
        yield np.array([[diag[0], off_diag[0], off_diag[1]], [0, diag[1], off_diag[2]],
                        [0, 0, diag[2]]])


def all_integer_transform_matrices_per_diag(n):
    """
    Yield all the integer transform matrices
    """
    diags = filterfalse(lambda x: x[0] * x[1] * x[2] != n, product(range(n + 1), repeat=3))

    for d in diags:
        yield all_integer_transform_matrices_given_diag(d)


def all_integer_transform_matrices(n):
    return chain(*all_integer_transform_matrices_per_diag(n))


def rotate_cells(cell, target_cell):
    """
    Rotate the cells such that one vector is parallel.
    And the other vector lies in a given plane
    """
    dot_prod_cell = cell.dot(cell.T)
    dot_prod_target_cell = target_cell.dot(target_cell.T)

    # Calculate unit vector cells
    uvec_cell = np.zeros_like(cell)
    uvec_target_cell = np.zeros_like(target_cell)
    for i in range(3):
        uvec_cell[i, :] = cell[i, :] / np.sqrt(dot_prod_cell[i, i])
        uvec_target_cell[i, :] = target_cell[i, :] / \
            np.sqrt(dot_prod_target_cell[i, i])

    # Rotate one vector to be parallel
    v = np.cross(uvec_cell[0, :], uvec_target_cell[0, :])
    c = uvec_cell[0, :].dot(uvec_target_cell[0, :])
    v_cross = np.zeros((3, 3))
    v[0, 1] = -v[2]
    v[0, 2] = v[1]
    v[1, 0] = v[2]
    v[1, 2] = -v[0]
    v[2, 0] = -v[1]
    v[2, 1] = v[0]
    R = np.eye(3) + v_cross + v_cross.dot(v_cross) / (1 + c)
    target_cell = R.dot(target_cell)
    return target_cell


def species_chempot2eci(bf_list, species_chempot):
    """
    Convert chemical potentials given for species to their corresponding
    singlet ECI values.

    Parameters:

    bf_list: list
        List of basis function values for each species

    species_chempot: dict
        Dictionary containing the chemical potential for each species. Note
        that the chemical potential for one species can be set to 0.0 without
        any loss of generality. Hence, the species_chempot should contain
        chemical potentials for all elements except one.
    """
    if len(species_chempot) != len(bf_list):
        msg = "Inconsistent number of chemical potentials. Basis functions\n"
        msg += f"{bf_list}. Passed chemical potentials {species_chempot}"
        raise ValueError(msg)

    n = len(species_chempot)
    mat = np.zeros((n, n))
    rhs = np.zeros(n)
    row = 0
    for sp, mu in species_chempot.items():
        for col, bf in enumerate(bf_list):
            mat[row, col] = bf[sp]
        rhs[row] = mu
        row += 1
    try:
        eci_chem_pot = np.linalg.solve(mat, rhs)
    except np.linalg.LinAlgError:
        inv_mat = np.linalg.pinv(mat)
        eci_chem_pot = inv_mat.dot(rhs)
    eci_chem_pot = eci_chem_pot.tolist()
    eci_dct = {f"c1_{i}": v for i, v in enumerate(eci_chem_pot)}
    return eci_dct


def bf2matrix(bfs):
    """
    Convert a list of basis functions to a matrix. Each column represents
    a species and each row represent a basis function.

    Parameter:

    bfs: list of dict
        List of dictionaries containing the basis function values
    """
    nrows = len(bfs)
    ncols = len(bfs[0])
    mat = np.zeros((nrows, ncols))
    keys = sorted(list(bfs[0].keys()))
    for i, bf in enumerate(bfs):
        for j, symb in enumerate(keys):
            mat[i, j] = bf[symb]
    return mat


def singlets2conc(bf_list, singlets):
    """
    Convert singlets to concentrations.

    Parameters:

    bf_list: list
        List with the basis functions (e.g [{'Au': 1, 'Cu': -1.0}])

    singlets: np.ndarray
        Array with singlets (NxM), M has to match the length of bf_list.
        The columns are assumed to given in sorted order (i.e. c1_0, c1_1,
        c1_2 etc.)
    """
    mat = bf2matrix(bf_list)

    # Add row to force that concentrations sum to 1
    mat = np.vstack((mat, np.ones(mat.shape[1])))
    singlets = np.hstack((singlets, np.ones((singlets.shape[0], 1))))

    res = np.linalg.solve(mat, singlets.T).T
    symbs = sorted(list(bf_list[0].keys()))
    concs = []
    for i in range(res.shape[0]):
        concs.append({symbs[j]: res[i, j] for j in range(res.shape[1])})
    return concs


def rate_bf_subsets(elems, bfs):
    """
    Rate different combinations of basis function according to how
    well it is able to distinguish the elements in the basis.

    Example:
    bfs = [{'Li': 1.0, 'X': 0.0, 'V': 0.0, 'F': 0.0},
           {'Li': 0.0, 'X': 1.0, 'V': 0.0, 'F': 0.0},
           {'Li': 0.0, 'X': 0.0, 'V': 1.0, 'F': 0.0}]

    If one wants basis functions for the triplet [Li, X, F] we need two
    basis functions. The following combinations are possible
    (bfs[0], bfs[1]), (bfs[0], bfs[2]), (bfs[1], bfs[2])

    The score is defined as the sum of the absolute value of the difference
    between the basis function value for the selected symbols. The score for
    the first combinations is thus

    score1 = |bfs[0]['Li'] - bfs[0]['F']| + |bfs[0]['Li'] - bfs[0]['X']| +
             |bfs[0]['F'] - bfs[0]['X']|  + |bfs[1]['Li'] - bfs[1]['F']| +
             |bfs[1]['Li'] - bfs[1]['X']| + |bfs[1]['F']  - bfs[1]['X']|

    Therefore,
    score1 = 1.0 + 1.0 + 0.0 + 0.0 + 1.0 + 1.0 = 4.0

    Parameter:

    bfs: list of dict
        List with dictionaries with the basis function values
    """
    score_and_comb = []
    for comb in combinations(range(len(bfs)), r=len(elems) - 1):
        chosen = [{s: bfs[x][s] for s in elems} for x in comb]
        mat = bf2matrix(chosen)
        score = 0.0
        for row in range(mat.shape[0]):
            for i in product(range(mat.shape[1]), repeat=2):
                score += abs(mat[row, i[0]] - mat[row, i[1]])

        score_and_comb.append((score, list(comb)))
    score_and_comb.sort(reverse=True)
    return score_and_comb


def select_bf_subsets(basis_elems, bfs):
    """
    Select a subset of basis functions that best describes the basis.
    The best subset is the one that yields the highest sum of the scores for
    each sublattice. For definition of score see docstring
    of `rate_bf_subsets`

    Parameters:

    basis_elem: nested list of strings
        Basis elements (e.g. [[Li, V], [O]])

    bfs: list of dicts
        List of dictionaries holding the basis function values
        (e.g. [{'Li' : 1.0, 'V': 0.0, 'O': 0.0},
               {'Li': 0.0, 'O': 1.0, 'V': 0.0}])
    """
    rates = []
    for elems in basis_elems:
        rates.append(rate_bf_subsets(elems, bfs))

    # Find the combination of subset selections that gives the overall highest
    # score given that a basis function can be present in only one basis
    best_selection = []
    best_score = None
    for comb in product(*rates):
        total_score = sum(rate[0] for rate in comb)
        selection = [rate[1] for rate in comb]

        # Check that no basis function is selected twice
        selected_bfs = set()
        duplicates = False
        for s in selection:
            for bf in s:
                if bf in selected_bfs:
                    duplicates = True
                else:
                    selected_bfs.add(bf)

        # Add penalty to the ones that has duplicate CF functions. This way we
        # select a combination that has the same basis function in multiple
        # atomic basis if it is possible
        if duplicates:
            total_score -= 1000.0

        if best_score is None or total_score > best_score:
            best_score = total_score
            best_selection = selection
    return best_selection


# pylint: disable=too-many-branches
def cname_lt(cname1, cname2):
    """
    Compare two cluster names to check if the first cluster name is
    smaller than (less than, or lt) the second cluster name. Since the cluster
    names take a form 'c#_d####_#', the prefix ('c#_d####') is evaluated as a
    string while the suffix ('#') is evaluated as an integer.

    Return `True` if cname1 < cname2 and `False` otherwise.
    """
    if not isinstance(cname1, str) and isinstance(cname2, str):
        raise TypeError('cnames should be strings.')

    if cname1 in ('c0', 'c1'):
        prefix1 = cname1
    else:
        prefix1 = cname1.rpartition("_")[0]

    if cname2 in ('c0', 'c1'):
        prefix2 = cname2
    else:
        prefix2 = cname2.rpartition("_")[0]

    if prefix1 < prefix2:
        return True
    if prefix1 > prefix2:
        return False

    # Case where prefixes are the same.
    if cname1 in ('c0', 'c1'):
        suffix1 = 0
    else:
        suffix1 = int(cname1.rpartition("_")[-1])

    if cname2 in ('c0', 'c1'):
        suffix2 = 0
    else:
        suffix2 = int(cname2.rpartition("_")[-1])

    if suffix1 < suffix2:
        return True
    return False


def aic(mse, num_features, num_data_points):
    """
    Return Afaike's information criteria

    Parameters:
    mse: float
        Mean square error

    num_features: int
        Number of features in the model

    num_data_points: int
        Number of data points
    """
    return 2.0 * num_features + num_data_points * np.log(mse)


def aicc(mse, num_features, num_data_points):
    """
    Return the modified Afaike's information criterion

    Parameters:
    mse: float
        Mean square error

    num_features: int
        Number of features in the model

    num_data_points: int
        Number of data points
    """
    if num_features >= num_data_points - 1:
        denum = 1.0
    else:
        denum = num_data_points - num_features - 1
    corr = (2 * num_features**2 + 2 * num_features) / denum
    return aic(mse, num_features, num_data_points) + corr


def bic(mse, num_features, num_data_points):
    """
    Return Bayes Information Criteria

    Parameters:
    mse: float
        Mean square error

    num_features: int
        Number of features

    num_data_points: int
        Number of data points
    """
    return np.log(num_data_points) * num_features + num_data_points * np.log(mse)


def get_extension(fname: Union[str, Path]) -> str:
    """
    Return the file extension of a filename

    Parameter:

    fname: str
        Filename
    """
    return Path(fname).suffix


def add_file_extension(fname: Union[str, Path], ext: str) -> str:
    """
    Adds the wanted file extension to a filename. If a file extension
    is already present and it matches the wanted file extension, nothing
    is done. If it does not match, a ValueError is raised. Finally, if
    no file extension exist the wanted extension is added

    :param fname: file name
    :param ext: extension (with .) example (.csv, .txt, .json)
    """
    fname = Path(fname)
    current_ext = fname.suffix
    if current_ext == ext:
        return str(fname)
    if current_ext == '':
        return str(fname.with_suffix(ext))
    raise ValueError(f"Passed extenstion {current_ext} expected {ext}")


def sort_cf_names(cf_names: tIterable[str]) -> List[str]:
    """
    Return a sorted list of correlation function names. The names are
    sorted according to the following criteria

    1. Size
    2. Diameter
    3. Lexicographical order of the name itself
    """
    sizes = [int(n[1]) for n in cf_names]
    # Regular expression that extracts all digits after the occurence
    # of _d (e.g. c2_d0001_0_00 --> 0001)
    prog = re.compile("_d(\\d+)")
    dia_str = [prog.findall(n) for n in cf_names]
    dia = []
    for d in dia_str:
        if d:
            dia.append(int(d[0]))
        else:
            dia.append(0)

    sort_obj = list(zip(sizes, dia, cf_names))
    sort_obj.sort()
    return [s[-1] for s in sort_obj]


def get_ids(select_cond: List[tuple], db_name: str) -> List[int]:
    """
    Return ids in the datase that correspond to the passed selection.

    :param select_cond: ASE select conditions.

    :return: List of database IDs matching the select conditions.

    """
    keys, cmps = parse_selection(select_cond)
    db = connect(db_name)
    sql, args = db.create_select_statement(keys, cmps)

    # Extract the ids in the database that corresponds to select_cond
    sql = sql.replace('systems.*', 'systems.id')
    with connect(db_name) as db:
        con = db.connection
        cur = con.cursor()
        cur.execute(sql, args)
        ids = [row[0] for row in cur.fetchall()]
    ids.sort()
    return ids


class SQLCursor(Protocol):

    def execute(self, sql: str, placeholder: Tuple[str]) -> None:
        pass

    def fetchall(self) -> tuple:
        pass


def get_attribute(ids: List[int], cur: SQLCursor, key: str, table: str) -> list:
    """
    Retrieve the value of the given key for the rows with the given ID of
    the database entry.

    :param ids: list of IDs
    :param cur: cursor for the database
    :param key: name of the key
    :param table: name of the table
    """
    known_tables = ["text_key_values", "number_key_values"]

    if table not in known_tables:
        raise ValueError(f"Table has to be one of {known_tables}")

    sql = f'SELECT value, id FROM {table} WHERE key=?'
    id_set = set(ids)
    cur.execute(sql, (key,))

    row_id_value = {}
    for value, row_id in cur.fetchall():
        if row_id in id_set:
            row_id_value[row_id] = value

    # Convert to a list that matches the order of the IDs that
    # was passed
    return [row_id_value[k] for k in ids]


def common_cf_names(ids: Set[int], cur: SQLCursor, table: str) -> Set[str]:
    """
    Extracts all correlation function names that are present for all
    ids

    :param ids: List of ids that should be checked
    :param cur: SQL cursor
    :param table: Table to check
    """
    known_tables = ['polynomial_cf', 'binary_linear_cf', 'trigonometric_cf']
    if table not in known_tables:
        raise ValueError(f"Table has to be one of {known_tables}")

    sql = f"SELECT key, id FROM {table}"
    cur.execute(sql)
    cf_names = {}
    for cf_name, row_id in cur.fetchall():
        if row_id in ids:
            current = cf_names.get(row_id, set())
            current.add(cf_name)
            cf_names[row_id] = current

    # Calculate the intersection between all sets
    return set.intersection(*list(cf_names.values()))


def constraint_is_redundant(A_lb: np.ndarray,
                            b_lb: np.ndarray,
                            c_lb: np.ndarray,
                            d: float,
                            A_eq: np.ndarray = None,
                            b_eq: np.ndarray = None) -> bool:
    """
    The method considers the following system

    min c.dot(x) for some arbitrary c

    subject to

    A_lb.dot(x) >= b_lb
    A_eq.dot(x) = b_eq

    If the additional constraint c_lb.dot(x) >= d is redundant, the method
    returns True. The constraint specified by c_lb is redundant if the solution
    to min c_lb.dot(x) subject to the constraint above satisfies
    c_lb.dot(x) >= d. This method is know as the Linear Programming Method.

    :param A_lb: Matrix specifying the lower bounds
    :param b_lb: Vector specifying the right hand side of the lower bound
        constraint.
    :param c_lb: Vector specifying the additional in-equality constraint
    :param d: Right hand side of the additional in-equality
    :param A_eq: Matrix specifying the equality constraint. If None,
        no equality constraints exist.
    :param b_eq: Vector specifuing the right hand side of the equality
        constraints. If None, no equality constraints exist
    """
    # Scipy uses upper bounds in stead of lower bounds, convert lower bounds
    # to upper bounds by changing the sign
    res = linprog(c_lb, A_ub=-A_lb, b_ub=-b_lb, A_eq=A_eq, b_eq=b_eq)
    return c_lb @ res.x >= d


def remove_redundant_constraints(A_lb: np.ndarray,
                                 b_lb: np.ndarray,
                                 A_eq: np.ndarray = None,
                                 b_eq: np.ndarray = None) -> Tuple[np.ndarray, np.ndarray]:
    """
    Remove all redundant constraints from A_lb and b_lb.

    min c.dot(x) for some arbitrary c

    subject to

    A_lb.dot(x) >= b_lb
    A_eq.dot(x) = b_eq

    :param A_lb: Matrix specifying the lower bounds
    :param b_lb: Vector specifying the right hand side of the lower bound
        constraint.
    :param A_eq: Matrix specifying the equality constraint. If None,
        no equality constraints exist.
    :param b_eq: Vector specifuing the right hand side of the equality
        constraints. If None, no equality constraints exist
    :return:
        A_lb, b_lb with only non-redundant constraints
    """
    redundant = []
    perturb = 1
    for i in range(A_lb.shape[0]):
        c_lb = A_lb[i, :]
        d = b_lb[i]

        # Make the constraint under consideration more generous by lowering the bound
        b_lb[i] -= perturb

        if constraint_is_redundant(A_lb, b_lb, c_lb, d, A_eq, b_eq):
            redundant.append(i)

        # Set the constraint back to the original value
        b_lb[i] += perturb
    return np.delete(A_lb, redundant, axis=0), np.delete(b_lb, redundant)


def remove_redundant_equations(A, b, tol=1e-6):
    R_trimmed = []
    indices = []

    Q, R = np.linalg.qr(A.T)

    k = 0
    for i in range(0, R.shape[1]):
        if abs(R[k, i]) > tol:
            R_trimmed.append(R[:, i])
            indices.append(i)
            k += 1

        if k == R.shape[0]:
            break

    R_trimmed = np.array(R_trimmed).T
    A_trimmed = Q.dot(R_trimmed).T
    return A_trimmed.copy(), b[indices]


class SystemChange(NamedTuple):
    index: int
    old_symb: str
    new_symb: str



import numpy as np
from typing import Dict, List


class GramSchmidtMonimial(object):
    """
    Class generates an orthogonal basis for an arbitrary number of
    elements.

    If x denotes the fictitious spin variable. For a binary system x typically
    takes the values -1 and 1, for a ternary -1,0,1 and so on.

    The class starts out from the monomial basis
    {1, x, x^2, x^3, x^4,...,x^n}

    If y_i denotes the new orthogonal basis function they are computed by the
    Gram-Schmidt procedure
    y_0 = 1
    y_1 = x - <x,y_0>y_0, then y_1 = y_1/||y_1||
    y_2 = x^2 - <x^2,y_1>y_1 - <x^2,y_2>y_2, then y_2 = y_2/||y_2||
    and so on
    <x,y> denotes the inner product between to functions
    and ||x|| = sqrt(<x,x>) is the norm of the function

    Parameter:

    num_symbols:
        Number of unique symbols
    """

    def __init__(self, num_symbols: int):
        self.values = []
        if num_symbols % 2 == 1:
            highest = (num_symbols - 1) / 2
        else:
            highest = num_symbols / 2

        # Assign spin value for each element
        while highest > 0:
            self.values.append(highest)
            self.values.append(-highest)
            highest -= 1
        if num_symbols % 2 == 1:
            self.values.append(0)

        self.bf_values = []

    def spin_dict(self, symbols: List[str]) -> Dict[str, int]:
        """
        Return a dictionary with the spin variable for each of the
        passed symbols
        """
        return {s: self.values[i] for i, s in enumerate(symbols)}

    def basis_functions(self, symbols: List[str]) -> List[Dict[str, float]]:
        """
        Construct the spin dictionary from a list of symbols
        """
        symbols = sorted(symbols)
        bf_list = []
        for bf in self.bf_values[1:]:
            bf_list.append({s: v for s, v in zip(symbols, bf)})
        return bf_list

    def build(self):
        """
        Populates the bf_values table
        The bf_value table has the following structure for 2 atoms
        -------------------------------------
        phi_0(values[0]) | phi_0(values[1]) |
        -------------------------------------
        phi_1(values[1]) | phi_1(values[1]) |
        -------------------------------------

        phi_i is the basis function derived by Gram-Schmidt from a monomial
        basis
        """
        # The first basis function is just the constant 1
        self.bf_values.append([1.0 for i in range(len(self.values))])

        # Loop over all basis functions
        for bf1 in range(1, len(self.values)):
            new_bf_values = []
            for i in range(len(self.values)):
                new_bf_values.append(self.evaluate_monomial_basis_function(bf1, self.values[i]))
                for bf2 in range(bf1):
                    new_bf_values[-1] -= self.dot_monomial_bf(bf1, bf2) \
                        * self.bf_values[bf2][i]

            self.bf_values.append(new_bf_values)

            # Normalize
            bf_norm = self.norm(bf1)
            for i in range(len(self.values)):
                self.bf_values[bf1][i] /= bf_norm

    def dot(self, bf1: int, bf2: int) -> float:
        """
        Performs the dot product between two basis functions

        Parameters:

        bf1:
            Index of the first basis function
        bf2:
            Index of the second basis function
        """
        if (bf1 >= len(self.values) or bf2 >= len(self.values)):
            raise ValueError("The provided indices has to be smaller "
                             "than total number of basis functions")

        dot_prod = 0.0
        for i in range(len(self.values)):
            dot_prod += self.bf_values[bf1][i] * self.bf_values[bf2][i]
        return dot_prod / len(self.values)

    def dot_monomial_bf(self, monomial: int, bf: int) -> float:
        """
        Perfmorms the dot product betweeen a monomial basis function and a
        "proper" basis function

        Parameters:

        monomial:
            Integer specifying the power of the monimial

        bf:
            Integer specifying the orthonormal basis function
        """
        dot_prod = 0.0
        for i in range(len(self.values)):
            dot_prod += self.evaluate_monomial_basis_function(
                monomial, self.values[i]) * self.bf_values[bf][i]
        return dot_prod / len(self.values)

    def norm(self, bf: List[float]) -> float:
        """
        Computes the norm of a basis function

        Parameter:

        bf:
            List with all the basis function values
        """
        return np.sqrt(self.dot(bf, bf))

    def eval(self, bf: List[float], valueIndex: int) -> float:
        """
        Evaluates the basis function

        Parameter:

        bf:
            List with all basis functions

        valueIndex:
            Index into the array of basis function values for basis function
            determined by bf
        """
        return self.bf_values[bf][valueIndex]

    def evaluate_monomial_basis_function(self, power: int, value: float) -> float:
        """
        Evaluate monimial basis

        Paraeters:

        power:
            Power of the monomial

        value:
            Value to be raised to the given power
        """
        return value**power



#!/usr/bin/env python
import os
import argparse
from .db import db_cli


# For Windows dependencies see
# https://stackoverflow.com/questions/40769386/kivy-windows-unable-to-find-any-valuable-window-provider-at-all
def install_gui_dependencies(args):
    import subprocess
    subprocess.check_call(['pip', 'install', 'kivy'] + args)

    if os.name == 'nt':
        # Windows dependencies
        pckgs = 'docutils pygments pypiwin32 kivy.deps.sdl2 kivy.deps.glew'
        subprocess.check_call(['pip', 'install'] + pckgs.split() + args)

    url = 'https://github.com/kivy-garden/graph/archive/master.zip'
    subprocess.check_call(['pip', 'install', url] + args)


def main():
    parser = argparse.ArgumentParser(description="CLEASE CLI")
    subparsers = parser.add_subparsers(help="Sub command help", dest="command")
    gui_parser = subparsers.add_parser("gui", help="Launches the CLEASE GUI")
    gui_parser.add_argument("--setup",
                            help="Install missing dependencies required to launch the "
                            "GUI",
                            action="store_true")

    db_parser = subparsers.add_parser("db", help="Launch the CLEASE DB CLI")
    db_parser.add_argument("name", help="Name of the database")
    db_parser.add_argument("--show",
                           help="[tab, names, cf]. If tab: The name of all correlation "
                           "function tables is shown.\n"
                           "If names, the name of all known correlation functions "
                           "is shown\n"
                           "If cf, the correlation functions of the given ID is "
                           "shown (See ID argument)")
    db_parser.add_argument("--id", help="Database ID to operate on")
    args = parser.parse_args()
    if args.command == 'gui':
        if args.setup:
            install_gui_dependencies([])
        else:
            from clease.gui.clease_gui import CleaseGUI
            CleaseGUI().run()
    elif args.command == 'db':
        db_cli(args)


if __name__ == "__main__":
    main()



from .cluster_fingerprint import *
from .cluster import *
from .cluster_generator import *
from .cluster_info_mapper import *
from .cluster_list import *
from .cluster_manager import *

from . import utils

ADDITIONAL = ('utils',)

__all__ = (cluster.__all__ + cluster_fingerprint.__all__ + cluster_generator.__all__ +
           cluster_info_mapper.__all__ + cluster_list.__all__ + cluster_manager.__all__ +
           ADDITIONAL)



"""Definition of ClusterExpansionSettings Class.

This module defines the base-class for storing the settings for performing
Cluster Expansion in different conditions.
"""
from copy import deepcopy
from typing import List, Dict, Optional, Union

import numpy as np
from ase.db import connect
from ase import Atoms

from clease.tools import wrap_and_sort_by_position
from clease.basis_function import BasisFunction
from clease.template_atoms import TemplateAtoms
from clease import AtomsManager
from clease.cluster import ClusterList, ClusterManager
from clease.template_filters import ValidConcentrationFilter
from clease.basis_function import Polynomial, Trigonometric, BinaryLinear
from . import Concentration

__all__ = ('ClusterExpansionSettings',)


class ClusterExpansionSettings:
    """Base class for all Cluster Expansion settings."""

    def __init__(self,
                 prim: Atoms,
                 concentration: Concentration,
                 size: Optional[int] = None,
                 supercell_factor: Optional[int] = 27,
                 db_name: str = 'clease.db',
                 max_cluster_size: int = 4,
                 max_cluster_dia: Union[float, List[float]] = 5.0) -> None:
        self.kwargs = {
            'size': size,
            'supercell_factor': supercell_factor,
            'db_name': db_name,
            'max_cluster_size': max_cluster_size,
            'max_cluster_dia': deepcopy(max_cluster_dia)
        }
        self._include_background_atoms = False
        self.trans_matrix = None
        self.concentration = self._get_concentration(concentration)
        self.cluster_list = ClusterList()
        self.basis_elements = deepcopy(self.concentration.basis_elements)
        self._check_first_elements()
        self.db_name = db_name
        self.size = to_3x3_matrix(size)

        self.prim_cell = prim
        self._tag_prim_cell()
        self._store_prim_cell()

        self.cluster_mng = ClusterManager(self.prim_no_bkg())

        prim_mng = AtomsManager(prim)
        prim_ind_by_basis = prim_mng.index_by_symbol([x[0] for x in self.basis_elements])
        conc_filter = ValidConcentrationFilter(concentration, prim_ind_by_basis)

        self.template_atoms = TemplateAtoms(self.prim_cell,
                                            supercell_factor=supercell_factor,
                                            size=self.size,
                                            skew_threshold=40,
                                            filters=[conc_filter])

        self.atoms_mng = AtomsManager(None)

        self.max_cluster_size = max_cluster_size
        self.max_cluster_dia = self._format_max_cluster_dia(max_cluster_dia)
        self.cluster_mng.build(max_size=self.max_cluster_size, max_cluster_dia=self.max_cluster_dia)

        self.set_active_template(atoms=self.template_atoms.weighted_random_template())

        unique_element_no_bkg = self.unique_element_without_background()
        self._basis_func_type = Polynomial(unique_element_no_bkg)

        if len(self.basis_elements) != self.num_basis:
            raise ValueError("list of elements is needed for each basis")

    @property
    def atoms(self) -> Atoms:
        return self.atoms_mng.atoms

    @property
    def all_elements(self) -> List[str]:
        return sorted([item for row in self.basis_elements for item in row])

    @property
    def num_elements(self) -> int:
        return len(self.all_elements)

    @property
    def unique_elements(self) -> List[str]:
        return sorted(list(set(deepcopy(self.all_elements))))

    @property
    def num_unique_elements(self) -> int:
        return len(self.unique_elements)

    @property
    def ref_index_trans_symm(self) -> List[int]:
        return [i[0] for i in self.index_by_sublattice]

    @property
    def skew_threshold(self):
        return self.template_atoms.skew_threshold

    @skew_threshold.setter
    def skew_threshold(self, threshold: int) -> None:
        '''
        Maximum acceptable skew level (ratio of max and min diagonal of the
        Niggli reduced cell)
        '''
        self.template_atoms.skew_threshold = threshold

    @property
    def background_indices(self) -> List[int]:
        """Get indices of the background atoms."""
        # check if any basis consists of only one element type
        basis = [i for i, b in enumerate(self.basis_elements) if len(b) == 1]

        bkg_indices = []
        for b_indx in basis:
            bkg_indices += self.index_by_basis[b_indx]
        return bkg_indices

    @property
    def include_background_atoms(self) -> bool:
        return self._include_background_atoms

    @include_background_atoms.setter
    def include_background_atoms(self, value: bool) -> None:
        if value == self._include_background_atoms:
            return
        self._include_background_atoms = value
        if self._include_background_atoms:
            self.cluster_mng = ClusterManager(self.prim_cell)
        else:
            self.cluster_mng = ClusterManager(self.prim_no_bkg())

        self.cluster_mng.build(max_size=self.max_cluster_size, max_cluster_dia=self.max_cluster_dia)
        self.create_cluster_list_and_trans_matrix()
        self.basis_func_type.unique_elements = \
            self.unique_element_without_background()

    @property
    def spin_dict(self) -> Dict[str, float]:
        return self.basis_func_type.spin_dict

    @property
    def basis_functions(self):
        return self.basis_func_type.basis_functions

    @property
    def ignore_background_atoms(self) -> bool:
        return not self.include_background_atoms

    @property
    def multiplicity_factor(self) -> Dict[str, float]:
        """Return the multiplicity factor of each cluster."""
        num_sites_in_group = [len(x) for x in self.index_by_sublattice]
        return self.cluster_list.multiplicity_factors(num_sites_in_group)

    @property
    def all_cf_names(self) -> List[str]:
        num_bf = len(self.basis_functions)
        return self.cluster_list.get_all_cf_names(num_bf)

    @property
    def num_cf(self) -> int:
        """Return the number of correlation functions."""
        return len(self.all_cf_names)

    @property
    def index_by_basis(self) -> List[List[int]]:
        first_symb_in_basis = [x[0] for x in self.basis_elements]
        return self.atoms_mng.index_by_symbol(first_symb_in_basis)

    @property
    def index_by_sublattice(self) -> List[List[int]]:
        return self.atoms_mng.index_by_tag()

    @property
    def num_basis(self) -> int:
        return len(self.basis_elements)

    @property
    def basis_func_type(self):
        return self._basis_func_type

    @basis_func_type.setter
    def basis_func_type(self, bf_type):
        """
        Type of basis function to use.
        It should be one of "polynomial", "trigonometric" or "binary_linear"
        """
        unique_element = self.unique_element_without_background()

        if isinstance(bf_type, BasisFunction):
            if bf_type.unique_elements != sorted(unique_element):
                raise ValueError("Unique elements in BasisFunction instance "
                                 "is different from the one in settings")
            self._basis_func_type = bf_type
        elif isinstance(bf_type, str):
            if bf_type.lower() == 'polynomial':
                self._basis_func_type = Polynomial(unique_element)
            elif bf_type.lower() == 'trigonometric':
                self._basis_func_type = Trigonometric(unique_element)
            elif bf_type.lower() == "binary_linear":
                self._basis_func_type = BinaryLinear(unique_element)
            else:
                msg = f"basis function type {bf_type} is not supported."
                raise ValueError(msg)
        else:
            raise ValueError("basis_function has to be an instance of BasisFunction or a string")

    def to_dict(self) -> Dict:
        return {
            'kwargs': self.kwargs,
            'include_background_atoms': self.include_background_atoms,
            'skew_threshold': self.skew_threshold,
            'basis_func_type': self.basis_func_type.todict()
        }

    def _get_concentration(self, concentration):
        if isinstance(concentration, Concentration):
            conc = concentration
        elif isinstance(concentration, dict):
            conc = Concentration.from_dict(concentration)
        else:
            raise TypeError("concentration has to be either dict or instance of Concentration")
        self.kwargs["concentration"] = conc.to_dict()
        return conc

    def prim_no_bkg(self):
        """
        Return an instance of the primitive cell where the background indices
        has been removed
        """
        prim = self.prim_cell.copy()
        bg_syms = self.get_bg_syms()
        delete = []
        for atom in prim:
            if atom.symbol in bg_syms:
                delete.append(atom.index)

        delete.sort(reverse=True)
        for i in delete:
            del prim[i]
        return prim

    def get_bg_syms(self):
        """
        Return the symbols in the basis where there is only one element
        """
        return set(x[0] for x in self.basis_elements if len(x) == 1)

    def unique_element_without_background(self):
        """Remove backgound elements."""
        if self.include_background_atoms:
            bg_sym = set()
        else:
            bg_sym = self.get_bg_syms()

            # Remove bg_syms that are also present in basis with more than one
            # element
            for elems in self.basis_elements:
                if len(elems) == 1:
                    continue
                to_be_removed = set()
                for s in bg_sym:
                    if s in elems:
                        to_be_removed.add(s)

                bg_sym -= to_be_removed

        unique_elem = set()
        for x in self.basis_elements:
            unique_elem.update(x)
        return list(unique_elem - bg_sym)

    def prepare_new_active_template(self, template):
        """Prepare necessary data structures when setting new template."""
        self.size = template.info['size']
        self.atoms_mng.atoms = template

    def set_active_template(self, atoms=None):
        """Set a new template atoms object."""
        if atoms is not None:
            template = self.template_atoms.get_template_matching_atoms(atoms=atoms)
        else:
            template = self.template_atoms.weighted_random_template()

        template = wrap_and_sort_by_position(template)

        if atoms is not None:
            # Check that the positions of the generated template
            # matches the ones in the passed object
            atoms = wrap_and_sort_by_position(atoms)
            if not np.allclose(template.get_positions(), atoms.get_positions()):
                raise ValueError(f"Inconsistent positions. Passed object\n"
                                 f"{atoms.get_positions()}\nGenerated template"
                                 f"\n{template.get_positions()}")
        self.prepare_new_active_template(template)
        self.create_cluster_list_and_trans_matrix()

    def _tag_prim_cell(self):
        """
        Add a tag to all the atoms in the unit cell to track the sublattice.
        Tags are added such that that the lowest tags corresponds to "active"
        sites, while the highest tags corresponds to background sites. An
        example is a system having three sublattices that can be occupied by
        more than one species, and two sublattices that can be occupied by
        only one species. The tags 0, 1, 2 will then be assigned to the three
        sublattices that can be occupied by more than one species, and the two
        remaining lattices will get the tag 3, 4.
        """
        bg_sym = self.get_bg_syms()
        tag = 0

        # Tag non-background elements first
        for atom in self.prim_cell:
            if atom.symbol not in bg_sym:
                atom.tag = tag
                tag += 1

        # Tag background elements
        for atom in self.prim_cell:
            if atom.symbol in bg_sym:
                atom.tag = tag
                tag += 1

    def _store_prim_cell(self):
        """Store unit cell to the database."""
        db = connect(self.db_name)
        shape = self.prim_cell.get_cell_lengths_and_angles()
        for row in db.select(name='primitive_cell'):
            uc_shape = row.toatoms().get_cell_lengths_and_angles()
            if np.allclose(shape, uc_shape):
                return row.id

        uid = db.write(self.prim_cell, name='primitive_cell')
        return uid

    def _get_prim_cell(self):
        raise NotImplementedError("This function has to be implemented in in derived classes.")

    def _format_max_cluster_dia(self, max_cluster_dia):
        """Get max_cluster_dia in numpy array form"""
        # max_cluster_dia is list or array
        if isinstance(max_cluster_dia, (list, np.ndarray, tuple)):
            # Length should be either max_cluster_size+1 or max_cluster_size-1
            mcd = np.array(max_cluster_dia, dtype=float)
            if len(max_cluster_dia) == self.max_cluster_size + 1:
                for i in range(2):
                    mcd[i] = 0.
            elif len(max_cluster_dia) == self.max_cluster_size - 1:
                mcd = np.insert(mcd, 0, [0., 0.])
            else:
                raise ValueError("Invalid length for max_cluster_dia.")
        # max_cluster_dia is int or float
        elif isinstance(max_cluster_dia, (int, float)):
            mcd = np.ones(self.max_cluster_size - 1, dtype=float) * max_cluster_dia
            mcd = np.insert(mcd, 0, [0., 0.])
        # Case for *None* or something else
        else:
            raise TypeError("max_cluster_dia is of wrong type, got: {}".format(
                type(max_cluster_dia)))
        return mcd.round(decimals=3)

    def _get_atoms(self):
        """Create atoms with a user-specified size."""
        atoms = self.prim_cell.copy() * self.size
        return wrap_and_sort_by_position(atoms)

    def create_cluster_list_and_trans_matrix(self):
        at_cpy = self.atoms
        self.cluster_list = self.cluster_mng.info_for_template(at_cpy)
        self.trans_matrix = self.cluster_mng.translation_matrix(at_cpy)

    def view_clusters(self):
        """Display all clusters along with their names."""
        from ase.gui.gui import GUI
        from ase.gui.images import Images
        figures = self.cluster_mng.get_figures()
        images = Images()
        images.initialize(figures)
        gui = GUI(images, expr='')
        gui.show_name = True
        gui.run()

    def get_all_templates(self):
        """
        Return a list with all template atoms
        """
        return self.template_atoms.get_all_templates()

    def view_templates(self):
        """
        Display all templates in the ASE GUi
        """
        from ase.visualize import view
        view(self.get_all_templates())

    def _check_first_elements(self):
        basis_elements = self.basis_elements
        num_basis = self.num_basis
        # This condition can be relaxed in the future
        first_elements = []
        for elements in basis_elements:
            first_elements.append(elements[0])
        if len(set(first_elements)) != num_basis:
            raise ValueError("First element of different basis should not be the same.")

    def save(self, filename):
        """Write Setting object to a file in JSON format.

        Parameters:

        filename: str
            Name of the file to store the necessary settings to initialize
            the Cluster Expansion calculations.
        """

        import json
        with open(filename, 'w') as outfile:
            json.dump(self.to_dict(), outfile, indent=2)


def to_3x3_matrix(size):
    if size is None:
        return None
    if isinstance(size, np.ndarray):
        size = size.tolist()

    # Is already a matrix
    if np.array(size).shape == (3, 3):
        return size

    if np.array(size).shape == (3,):
        return np.diag(size).tolist()

    raise ValueError(f"Cannot convert passed array {size} to 3x3 matrix")



from clease.calculator import attach_calculator
from clease.montecarlo import SGCMonteCarlo, MetaDynamicsSampler
from clease.gui.meta_dyn_gui_updater import MetaDynGuiUpdater


class MetaDynRunner(object):

    def __init__(self,
                 atoms=None,
                 meta_page=None,
                 max_sweeps=None,
                 settings=None,
                 app_root=None,
                 eci=None,
                 status=None,
                 mc_params=None,
                 T=None,
                 bias=None,
                 flat=None,
                 mod_factor=None,
                 backup_file=None,
                 update_interval=10):
        self.atoms = atoms
        self.T = T
        self.meta_page = meta_page
        self.update_interval = update_interval
        self.max_sweeps = max_sweeps
        self.settings = settings
        self.app_root = app_root
        self.eci = eci
        self.orig_template = settings.atoms.copy()
        self.status = status
        self.mc_params = mc_params
        self.bias = bias
        self.flat = flat
        self.mod_factor = mod_factor
        self.backup_file = backup_file

    def _attach_calc(self):
        self.status.text = 'Attaching calculator...'

        # Temporarily disable info update during initalisation
        self.app_root.info_update_disabled = True
        self.app_root.view_mc_cell_disabled = True
        self.atoms = attach_calculator(settings=self.settings, atoms=self.atoms, eci=self.eci)
        self.app_root.active_template_is_mc_cell = True
        self.app_root.info_update_disabled = False
        self.app_root.view_mc_cell_disabled = False

    def run(self):
        self.meta_page.mc_is_running = True
        update_status_on_end = True
        self._attach_calc()

        try:
            run_calc = True
            self.status.text = 'Running metadynamics MC...'
            if self.mc_params['name'] == 'Semi-grand canonical':
                mc = SGCMonteCarlo(self.atoms, self.T, symbols=self.mc_params['symbols'])
            else:
                msg = f"Unknown MC algorithm. Params: {self.mc_params}"
                self.status.text = msg
                run_calc = False

            if run_calc:
                meta_sampler = MetaDynamicsSampler(mc=mc,
                                                   bias=self.bias,
                                                   mod_factor=self.mod_factor,
                                                   fname=self.backup_file)

                updater = MetaDynGuiUpdater(meta_dyn_page=self.meta_page,
                                            meta_dyn_sampler=meta_sampler)
                meta_sampler.add_observer(updater, interval=self.update_interval)
                self.meta_page.bind_meta_dyn_sampler(meta_sampler)
                meta_sampler.run(max_sweeps=self.max_sweeps)
                self.meta_page.detach_meta_dyn_sampler()
        except Exception as exc:
            self.status.text = str(exc)
            update_status_on_end = False

        # Reset the old template
        self.settings.set_active_template(atoms=self.orig_template)
        self.app_root.active_template_is_mc_cell = False
        self.meta_page.mc_is_running = False

        if update_status_on_end:
            msg = 'MC finished. Lower mod. factor and rerun to improve the '
            msg += 'result'
            self.status.text = msg



"""Definitions of Cluster Expansion settings for bulk.

Cluster Expansion can be peformed on bulk material using either CEBulk
or CECrystal class.
"""
from ase.build import bulk
from ase.spacegroup import crystal
from clease.tools import wrap_and_sort_by_position
from .settings import ClusterExpansionSettings
from copy import deepcopy

__all__ = (
    'CEBulk',
    'CECrystal',
)


def CEBulk(concentration,
           crystalstructure='sc',
           a=None,
           c=None,
           covera=None,
           u=None,
           size=None,
           supercell_factor=27,
           db_name='clease.db',
           max_cluster_size=4,
           max_cluster_dia=(5.0, 5.0, 5.0)):
    """
    Specify cluster expansion settings for bulk materials defined based on
    crystal structures.

    Parameters:

    concentration: Concentration object or dict
        Concentration object or dictionary specifying the basis elements and
        concentration range of constituting species

    crystalstructure: str
        Must be one of sc, fcc, bcc, hcp, diamond, zincblende, rocksalt,
        cesiumchloride, fluorite or wurtzite.

    a: float
        Lattice constant.

    c: float
        Lattice constant.

    covera: float
        c/a ratio used for hcp.  Default is ideal ratio: sqrt(8/3).

    u: float
        Internal coordinate for Wurtzite structure.

    size: list
        Size of the supercell (e.g., [2, 2, 2] for 2x2x2 cell).
        `supercell_factor` is ignored if both `size` and `supercell_factor`
        are specified.

    supercell_factor: int
        Maximum multipilicity factor for limiting the size of supercell
        created from the primitive cell. `supercell_factor` is ignored if
        both `size` and `supercell_factor` are specified.

    db_name: str
        Name of the database file

    max_cluster_size: int
        Maximum size (number of atoms in a cluster)

    max_cluster_dia: list of int or float
        A list of int or float containing the maximum diameter of clusters
        (in Å)
    """
    structures = {
        'sc': 1,
        'fcc': 1,
        'bcc': 1,
        'hcp': 1,
        'diamond': 1,
        'zincblende': 2,
        'rocksalt': 2,
        'cesiumchloride': 2,
        'fluorite': 3,
        'wurtzite': 2
    }

    num_basis = len(concentration.orig_basis_elements)
    if num_basis != structures[crystalstructure]:
        msg = f"{crystalstructure} has {structures[crystalstructure]} basis. "
        msg += "The number of basis specified by basis_elements is "
        msg += f"{num_basis}"
        raise ValueError(msg)

    basis_elements = concentration.orig_basis_elements
    name = ''.join(x[0] for x in basis_elements)
    prim = bulk(name=name, crystalstructure=crystalstructure, a=a, c=c, covera=covera, u=u)
    prim = wrap_and_sort_by_position(prim)

    settings = ClusterExpansionSettings(prim, concentration, size, supercell_factor, db_name,
                                        max_cluster_size, max_cluster_dia)

    settings.kwargs.update({
        'crystalstructure': crystalstructure,
        'a': a,
        'c': c,
        'covera': covera,
        'u': u,
        'factory': 'CEBulk'
    })
    return settings


def CECrystal(concentration,
              spacegroup=1,
              basis=None,
              cell=None,
              cellpar=None,
              ab_normal=(0, 0, 1),
              size=None,
              supercell_factor=27,
              db_name='clease.db',
              max_cluster_size=4,
              max_cluster_dia=(5.0, 5.0, 5.0)):
    """Store CE settings on bulk materials defined based on space group.

    Parameters:

    concentration: Concentration object or dict
        Concentration object or dictionary specifying the basis elements and
        concentration range of constituting species

    spacegroup: int | string | Spacegroup instance
        Space group given either as its number in International Tables
        or as its Hermann-Mauguin symbol.

    basis: list of scaled coordinates
        Positions of the unique sites corresponding to symbols given
        either as scaled positions or through an atoms instance.

    cell: 3x3 matrix
        Unit cell vectors.

    cellpar: [a, b, c, alpha, beta, gamma]
        Cell parameters with angles in degree. Is not used when `cell`
        is given.

    ab_normal: vector
        Is used to define the orientation of the unit cell relative
        to the Cartesian system when `cell` is not given. It is the
        normal vector of the plane spanned by a and b.

    size: list
        Size of the supercell (e.g., [2, 2, 2] for 2x2x2 cell).
        `supercell_factor` is ignored if both `size` and `supercell_factor`
        are specified.

    supercell_factor: int
        Maximum multipilicity factor for limiting the size of supercell
        created from the primitive cell. `supercell_factor` is ignored if
        both `size` and `supercell_factor` are specified.

    db_name: str
        name of the database file

    max_cluster_size: int
        maximum size (number of atoms in a cluster)

    max_cluster_dia: list of int or float
        A list of int or float containing the maximum diameter of clusters
        (in Å)
    """

    symbols = []
    num_basis = len(concentration.orig_basis_elements)
    for x in range(num_basis):
        symbols.append(concentration.orig_basis_elements[x][0])

    prim = crystal(symbols=symbols,
                   basis=basis,
                   spacegroup=spacegroup,
                   cell=cell,
                   cellpar=cellpar,
                   ab_normal=ab_normal,
                   size=[1, 1, 1],
                   primitive_cell=True)
    prim = wrap_and_sort_by_position(prim)

    settings = ClusterExpansionSettings(prim, concentration, size, supercell_factor, db_name,
                                        max_cluster_size, max_cluster_dia)
    settings.kwargs.update({
        'basis': deepcopy(basis),
        'spacegroup': spacegroup,
        'cell': cell,
        'cellpar': cellpar,
        'ab_normal': ab_normal,
        'factory': 'CECrystal'
    })
    return settings



"""Module for generating new structures."""
import os
import math
import time
import logging
from copy import deepcopy
import numpy as np
from numpy.random import choice
from numpy.linalg import inv, pinv
from ase.db import connect
from clease.settings import ClusterExpansionSettings
from clease.corr_func import CorrFunction
from clease.tools import wrap_and_sort_by_position
from clease.calculator import Clease
from clease.montecarlo import Montecarlo
from clease.montecarlo.observers import LowestEnergyStructure, Snapshot
from clease.montecarlo.constraints import ConstrainSwapByBasis
from clease.montecarlo.montecarlo import TooFewElementsError
from ase.io.trajectory import TrajectoryReader

logger = logging.getLogger(__name__)


class StructureGenerator:
    """Base class for generating new strctures."""

    def __init__(self,
                 settings,
                 atoms,
                 struct_per_gen,
                 init_temp=None,
                 final_temp=None,
                 num_temp=5,
                 num_steps_per_temp=10000):
        if not isinstance(settings, ClusterExpansionSettings):
            raise TypeError("settings must be CEBulk or CECrystal " "object")

        self.settings = settings
        self.trans_matrix = settings.trans_matrix
        self.cf_names = self.settings.all_cf_names
        self.corrFunc = CorrFunction(settings)
        self.cfm = self._get_full_cf_matrix()
        self.atoms = wrap_and_sort_by_position(atoms.copy())
        self.settings.set_active_template(atoms=self.atoms)
        if self._is_valid(atoms):
            if len(atoms) != len(settings.atoms):
                raise ValueError("Passed Atoms has a wrong size.")
        else:
            raise ValueError("concentration of the elements in the provided"
                             " atoms is not consistent with the settings.")

        # eci set to 1 to ensure that all correlation functions are included
        # but the energy produced from this should never be used
        self.eci = {name: 1. for name in self.cf_names}
        calc = Clease(self.settings, eci=self.eci)
        self.atoms.calc = calc
        self.output_every = 30
        self.init_temp = init_temp
        self.final_temp = final_temp
        self.temp = self.init_temp
        self.num_temp = num_temp
        self.num_steps_per_temp = num_steps_per_temp
        self.alter_composition = True

        # Structures and correlation function associated with
        # the generated structure so far
        self.generated_structure = None
        self.cf_generated_structure = None

    def _is_valid(self, atoms):
        return self.settings.concentration.is_valid(self.settings.index_by_basis, atoms)

    def _reset(self):
        pass

    def generate(self):
        """Generate new structures."""
        # Start
        self._reset()
        if self.init_temp is None or self.final_temp is None:
            self.init_temp, self.final_temp = self._determine_temps()

        if self.init_temp <= self.final_temp:
            raise ValueError("Initial temperature must be higher than final temperature")
        self._reset()

        temps = np.logspace(math.log10(self.init_temp), math.log10(self.final_temp), self.num_temp)
        now = time.time()
        change_element = self._has_more_than_one_conc()

        N = len(temps)
        # Number of integers for max counters
        len_num_steps = len(str(self.num_steps_per_temp))
        len_max_temp = len(str(N))
        logger.info("Generating log statements every %s s", self.output_every)

        for Ti, temp in enumerate(temps):
            self.temp = temp
            num_accepted = 0
            count = 0
            while count < self.num_steps_per_temp:
                count += 1

                if time.time() - now > self.output_every:
                    acc_rate = float(num_accepted) / count
                    msg = (
                        f"T Step: {Ti+1:>{len_max_temp}} of {N}, "
                        # Print same len as max required
                        f"Temp: {temp:>10.3f}, "
                        # Print same len as max required
                        f"MC Step: {count:>{len_num_steps}d} of "
                        f"{self.num_steps_per_temp}, "
                        f"Accept. rate: {acc_rate*100:.2f} %")
                    logger.info(msg)
                    now = time.time()

                if choice([True, False]) and self.alter_composition:
                    logger.debug('Change element type? %s', change_element)
                    # Change element Type
                    if change_element:
                        self._change_element_type()
                    else:
                        continue
                else:
                    indx = self._swap_two_atoms()
                    logger.debug('Swapping indices %s', indx)
                    if self.atoms[indx[0]].symbol == \
                            self.atoms[indx[1]].symbol:
                        logger.debug('Same symbols detected, skipping.')
                        continue
                self.atoms.get_potential_energy()

                if self._accept():
                    num_accepted += 1
                    self.atoms.calc.clear_history()
                else:
                    self.atoms.calc.restore()

        # Create a new calculator and attach it to the generated structure
        calc = Clease(self.settings, eci=self.eci, init_cf=self.cf_generated_structure)
        self.generated_structure.calc = calc

        # Force an energy evaluation to update the results dictionary
        self.generated_structure.get_potential_energy()

        # Check that correlation function match the expected value
        self._check_consistency()
        cf = self.corrFunc.get_cf(self.generated_structure)
        return self.generated_structure, cf

    def _set_generated_structure(self):
        self.generated_structure = self.atoms.copy()
        cf_dict = self.atoms.calc.get_cf()
        self.cf_generated_structure = deepcopy(cf_dict)

    def _accept(self):
        raise NotImplementedError('_accept should be implemented in the ' 'inherited class.')

    def _estimate_temp_range(self):
        raise NotImplementedError('_estimate_temp_range should be '
                                  'implemented in the inherited class.')

    def _optimal_structure(self):
        raise NotImplementedError('_optimal_structure shoud be implemented '
                                  'in the inherited class.')

    def _determine_temps(self):
        logger.info("Temperature range not given. Determining the range automatically.")
        self._reset()
        count = 0
        max_count = 100
        now = time.time()
        # To avoid errors, just set the temperature to an arbitrary file
        self.temp = 10000000.0
        while count < max_count:
            if time.time() - now > self.output_every:
                logger.info("Progress (%.3f %%)", 100 * count / max_count)
                now = time.time()

            if choice([True, False]) and self.alter_composition:
                # Change element Type
                if self._has_more_than_one_conc():
                    self._change_element_type()
                    count += 1
                else:
                    continue
            else:
                indx = self._swap_two_atoms()
                if self.atoms[indx[0]].symbol == \
                        self.atoms[indx[1]].symbol:
                    continue
                count += 1
            self.atoms.get_potential_energy()

            # By calling accept statistics on the correlation
            # function and variance will be collected
            self._accept()
        init_temp, final_temp = self._estimate_temp_range()
        self.temp = init_temp
        logger.info("init_temp=%s, final_temp=%s", init_temp, final_temp)
        return init_temp, final_temp

    def _swap_two_atoms(self):
        """Swap two randomly chosen atoms."""
        indx = np.zeros(2, dtype=int)
        symbol = [None] * 2

        basis_elements = self.settings.basis_elements
        num_basis = self.settings.num_basis

        # pick fist atom and determine its symbol and type
        while True:
            basis = choice(range(num_basis))
            # a basis with only 1 type of element should not be chosen
            if len(basis_elements[basis]) < 2:
                continue
            indx[0] = choice(self.settings.index_by_basis[basis])
            symbol[0] = self.atoms[indx[0]].symbol
            break
        # pick second atom that occupies the same basis.
        while True:
            indx[1] = choice(self.settings.index_by_basis[basis])
            symbol[1] = self.atoms[indx[1]].symbol
            if symbol[1] in basis_elements[basis]:
                break

        # Swap two elements
        self.atoms[indx[0]].symbol = symbol[1]
        self.atoms[indx[1]].symbol = symbol[0]

        return indx

    def _has_more_than_one_conc(self):
        ranges = self.settings.concentration.get_individual_comp_range()
        for r in ranges:
            if abs(r[1] - r[0]) > 0.01:
                return True
        return False

    def _change_element_type(self):
        """Change the type of element for the atom with a given index.

        If index and replacing element types are not specified, they are
        randomly generated.
        """
        basis_elements = self.settings.basis_elements
        num_basis = self.settings.num_basis
        # ------------------------------------------------------
        # Change the type of element for a given index if given.
        # If index not given, pick a random index
        # ------------------------------------------------------
        while True:
            basis = choice(range(num_basis))
            # a basis with only 1 type of element should not be chosen
            if len(basis_elements[basis]) < 2:
                continue

            indx = choice(self.settings.index_by_basis[basis])
            old_symbol = self.atoms[indx].symbol

            # change element type
            new_symbol = choice(basis_elements[basis])
            if new_symbol != old_symbol:
                self.atoms[indx].symbol = new_symbol

                if self.settings.concentration.is_valid(self.settings.index_by_basis, self.atoms):
                    break
                self.atoms[indx].symbol = old_symbol

    def _check_consistency(self):
        # Check to see if the cf is indeed preserved
        final_cf = \
            self.corrFunc.get_cf_by_names(
                self.generated_structure,
                self.atoms.calc.cf_names)
        for k in final_cf:
            if abs(final_cf[k] - self.cf_generated_structure[k]) > 1E-6:
                msg = 'Correlation function changed after simulated annealing'
                logger.error(msg)

                # Print a summary of all basis functions (useful for debuggin)
                logger.debug('Basis functions:')
                for k in final_cf:
                    logger.debug('   %s: %s, %s', k, final_cf[k], self.cf_generated_structure[k])
                raise ValueError(msg)

    def _get_full_cf_matrix(self):
        """Get correlation function of every entry in DB."""
        cfm = []
        db = connect(self.settings.db_name)
        tab_name = f"{self.settings.basis_func_type.name}_cf"
        for row in db.select(struct_type='initial'):
            cfm.append([row[tab_name][x] for x in self.cf_names])
        cfm = np.array(cfm, dtype=float)
        return cfm


class ProbeStructure(StructureGenerator):
    """Generate probe structures.

    Based on simulated annealing according to the recipe in
    PRB 80, 165122 (2009).

    Parameters:

    settings: CEBulk or BulkSapcegroup object

    atoms: Atoms object
        initial structure to start the simulated annealing

    struct_per_gen: int
        number of structures to be generated per generation

    init_temp: int or float
        initial temperature (does not represent *physical* temperature)

    final_temp: int or float
        final temperature (does not represent *physical* temperature)

    num_temp: int
        number of temperatures to be used in simulated annealing

    num_steps_per_temp: int
        number of steps per temperature in simulated annealing

    approx_mean_var: bool
        whether or not to use a spherical and isotropical distribution
        approximation scheme for determining the mean variance.
        -'True': Assume a spherical and isotropical distribution of
                 structures in the configurational space.
                 Corresponds to eq.4 in PRB 80, 165122 (2009)
        -'False': Use sigma and mu of eq.3 in PRB 80, 165122 (2009)
                  to characterize the distribution of structures in
                  population.
                  Requires pre-sampling of random structures before
                  generating probe structures.
                  Reads sigma and mu from 'probe_structure-sigma_mu.npz' file.
    """

    def __init__(self,
                 settings,
                 atoms,
                 struct_per_gen,
                 init_temp=None,
                 final_temp=None,
                 num_temp=5,
                 num_steps_per_temp=1000,
                 approx_mean_var=False):

        StructureGenerator.__init__(self, settings, atoms, struct_per_gen, init_temp, final_temp,
                                    num_temp, num_steps_per_temp)
        self.o_cf = self.atoms.calc.cf
        self.o_cfm = np.vstack((self.cfm, self.o_cf))
        self.approx_mean_var = approx_mean_var
        fname = 'probe_structure-sigma_mu.npz'
        if not approx_mean_var:
            if os.path.isfile(fname):
                data = np.load(fname)
                self.sigma = data['sigma']
                self.mu = data['mu']
                self.o_mv = mean_variance(self.o_cfm, self.sigma, self.mu)
            else:
                raise IOError(f"'{fname}' not found.")
        else:
            self.o_mv = mean_variance_approx(self.o_cfm)
        self.avg_mv = 0.0
        self.num_steps = 0
        self.avg_diff = 0.0

        self.min_mv = None

    def _accept(self):
        """Accept the last change."""
        cfm = np.vstack((self.cfm, self.atoms.calc.cf))
        if self.approx_mean_var:
            n_mv = mean_variance_approx(cfm)
        else:
            n_mv = mean_variance(cfm, self.sigma, self.mu)

        # Always accept the first move
        if self.generated_structure is None:
            self.min_mv = n_mv
            self._set_generated_structure()
            return True

        if n_mv < self.o_mv:
            self.min_mv = n_mv
            self._set_generated_structure()

        if self.o_mv > n_mv:
            accept = True
        else:
            accept = np.exp((self.o_mv - n_mv) / self.temp) > np.random.uniform()

        self.avg_diff += abs(n_mv - self.o_mv)
        if accept:
            self.o_mv = n_mv

        self.avg_mv += self.o_mv
        self.num_steps += 1
        return accept

    def _estimate_temp_range(self):
        if self.num_steps == 0:
            return 100000.0, 1.0
        avg_diff = self.avg_diff / self.num_steps
        init_temp = 10 * avg_diff
        final_temp = 0.01 * avg_diff
        return init_temp, final_temp

    @property
    def avg_mean_variance(self):
        if self.num_steps == 0:
            return 0.0
        return self.avg_mv / self.num_steps

    def _reset(self):
        self.avg_mv = 0.0
        self.avg_diff = 0.0
        self.num_steps = 0


class GSStructure(StructureGenerator):
    """Generate ground-state structure.

    Parameters:

    settings: CEBulk or BulkSapcegroup object

    atoms: Atoms object
        initial structure to start the simulated annealing

    struct_per_gen: int
        number of structures to be generated per generation

    init_temp: int or float
        initial temperature (does not represent *physical* temperature)

    final_temp: int or float
        final temperature (does not represent *physical* temperature)

    num_temp: int
        number of temperatures to be used in simulated annealing

    num_steps_per_temp: int
        number of steps per temperature in simulated annealing

    eci: dict
        Dictionary containing cluster names and their ECI values
    """

    def __init__(self,
                 settings,
                 atoms,
                 struct_per_gen,
                 init_temp=2000,
                 final_temp=10,
                 num_temp=10,
                 num_steps_per_temp=100000,
                 eci=None):
        StructureGenerator.__init__(self, settings, atoms, struct_per_gen, init_temp, final_temp,
                                    num_temp, num_steps_per_temp)
        self.alter_composition = False
        self.eci = eci
        calc = Clease(self.settings, eci=eci)
        self.atoms.calc = calc

    def generate(self):
        low_en_obs = LowestEnergyStructure(self.atoms, verbose=False)
        cnst = ConstrainSwapByBasis(self.atoms, self.settings.index_by_basis)

        temps = np.logspace(math.log10(self.init_temp), math.log10(self.final_temp), self.num_temp)
        try:
            for T in temps.tolist():
                logger.info("Current temperature: %.3f K", T)
                mc = Montecarlo(self.atoms, T)
                mc.attach(low_en_obs)
                mc.generator.add_constraint(cnst)
                mc.run(steps=self.num_steps_per_temp)
            self.generated_structure = low_en_obs.emin_atoms
        except TooFewElementsError:
            # In case there are two few elements, the ground state is equal to
            # the initial configuration
            self.generated_structure = self.atoms

        cf = self.corrFunc.get_cf(self.generated_structure)
        # cf_trimmed = {key: value for (key, value) in cf
        #               if key in list(self.eci.keys())}
        calc = Clease(self.settings, eci=self.eci)
        self.generated_structure.calc = calc
        self.generated_structure.get_potential_energy()

        return self.generated_structure, cf


class MetropolisTrajectory(object):

    def __init__(self, settings, atoms, num):
        self.settings = settings
        self.atoms = atoms
        self.num = num

    def generate(self):
        cnst = ConstrainSwapByBasis(self.atoms, self.settings.index_by_basis)
        eci = {'c0': 0.0}
        calc = Clease(self.settings, eci=eci)
        self.atoms.calc = calc

        mc = Montecarlo(self.atoms, 10000000)
        mc.generator.add_constraint(cnst)
        fname = 'metropolis_traj'
        obs = Snapshot(fname=fname, atoms=self.atoms)
        mc.attach(obs, interval=1)

        mc.run(steps=self.num)

        reader = TrajectoryReader(obs.fname)
        all_atoms = [a for a in reader]
        os.remove(obs.fname)
        return all_atoms


def mean_variance_full(cfm):
    prec = precision_matrix(cfm)
    mv = 0.
    for x in range(cfm.shape[0]):
        mv += cfm[x].dot(prec).dot(cfm[x].T)
    mv = mv / cfm.shape[0]
    return mv


def mean_variance(cfm, sigma, mu):
    prec = precision_matrix(cfm)
    return np.trace(prec.dot(sigma)) + mu.dot(prec).dot(mu.T)


def mean_variance_approx(cfm):
    prec = precision_matrix(cfm)
    return np.trace(prec)


def precision_matrix(cfm):
    try:
        prec = inv(cfm.T.dot(cfm))
    # if inverting matrix leads to a singular matrix, reduce the matrix
    except np.linalg.linalg.LinAlgError:
        prec = pinv(cfm.T.dot(cfm))
    return prec



from kivy.uix.screenmanager import Screen
from clease.gui.load_save_dialog import LoadDialog
from clease.gui.help_message_popup import HelpMessagePopup
from kivy.uix.popup import Popup
from kivy.app import App
import subprocess
from threading import Thread
from clease.gui.constants import JOB_EXEC_MSG


class JobExec(Screen):
    _pop_up = None

    def open_file_dialog(self):
        content = LoadDialog(load=self.load, cancel=self.dismiss_popup)
        self._pop_up = Popup(title="Load script",
                             content=content,
                             pos_hint={
                                 'right': 0.95,
                                 'top': 0.95
                             },
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def to_dict(self):
        """Return content as dict."""
        return {
            'cmd': self.ids.cmdInput.text,
            'script': self.ids.scriptInput.text,
            'dbIds': self.ids.dbIdsInput.text,
            'additionCmdArgs': self.ids.cmdArgsInput.text
        }

    def from_dict(self, dct):
        self.ids.cmdInput.text = dct.get('cmd', 'python')
        self.ids.scriptInput.text = dct.get('script', 'myscript.py')
        self.ids.dbIdsInput.text = dct.get('dbIds', '')
        self.ids.cmdArgsInput.text = dct.get('additionCmdArgs')

    def dismiss_popup(self):
        if self._pop_up is None:
            return
        self._pop_up.dismiss()
        self._pop_up = None

    def load(self, path, filename):
        self.db_path = path

        if len(filename) == 0:
            self.ids.scriptInput.text = path
        else:
            self.ids.scriptInput.text = filename[0]
        self.dismiss_popup()

    def show_help_msg(self):
        msg = 'The program will execute the following command:\n'
        msg += '<cmd> <script> dbID <arg1> <arg2> <arg3>...,\n'
        msg += 'where <cmd> and <script> are the two first entries on\n'
        msg += 'the page. dbID is an integer and <arg1>, <arg2> etc.\n'
        msg += 'are optional arguments given as a comma separated list\n'
        msg += 'If you give a comma separated list of IDs, the command \n'
        msg += 'will be launched for each ID. Furthermore, you can also\n'
        msg += 'specify a range of IDs like this 4-7 which is expanded to\n'
        msg += '4, 5, 6, 7.'
        content = HelpMessagePopup(close=self.dismiss_popup, msg=msg)
        self._pop_up = Popup(title="Calculation help",
                             content=content,
                             pos_hint={
                                 'right': 0.95,
                                 'top': 0.95
                             },
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def show_args_help(self):
        msg = 'In many cases, the IDs of the calculations are not sufficient\n'
        msg += 'and require more arguments to be passed. One example is that\n'
        msg += 'you want to perform an initial relaxation with lower energy\n'
        msg += 'cutoff and k-point density, followed by another calculation\n'
        msg += 'with higher energy cutoff and k-point density. You can \n'
        msg += 'specify additional arguments by passing comma-separated\n'
        msg += 'values. For instance, if your script can accept energy cutoff\n'
        msg += 'and k-point density as arguments, you can add 500, 5.4 to the\n'
        msg += 'field to pass the respective values. The executed command\n'
        msg += 'when the run button is pressed will be:\n\n'
        msg += 'python myscript.py <dBId> 500 5.4\n'
        content = HelpMessagePopup(close=self.dismiss_popup, msg=msg)
        self._pop_up = Popup(title="Additional args help",
                             content=content,
                             pos_hint={
                                 'right': 0.95,
                                 'top': 0.95
                             },
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def run_on_separate_thread(self, cmd, script, ids, args):
        app = App.get_running_app()
        for i, uid in enumerate(ids):
            msg = f"Running job {i} of {len(ids)}"
            app.root.ids.status.text = msg
            subprocess.check_call([cmd, script, str(uid)] + args)
        app.root.ids.status.text = JOB_EXEC_MSG['jobs_finished']

    def _resolve_id_ranges(self, dbIds):
        """Resolve ranges IDs."""
        ids = []
        for item in dbIds:
            if '-' in item:
                splitted = item.split('-')
                min_id = int(splitted[0])
                max_id = int(splitted[1])
                ids += list(range(min_id, max_id + 1))
            else:
                ids.append(int(item))
        return ids

    def run(self):
        args = self.ids.cmdArgsInput.text.split(',')
        app = App.get_running_app()
        try:
            dbIds = self.ids.dbIdsInput.text.split(',')
            dbIds = self._resolve_id_ranges(dbIds)
        except Exception:
            app.root.ids.status.text = JOB_EXEC_MSG['db_id_error']
            return

        cmd = self.ids.cmdInput.text
        script = self.ids.scriptInput.text
        Thread(target=self.run_on_separate_thread, args=(cmd, script, dbIds, args)).start()



from kivy.uix.popup import Popup
from clease.gui.load_save_dialog import LoadDialog
from clease import Evaluate
from clease.regression import GAFit, LinearRegression
from kivy.app import App
import json
from clease.gui.util import (parse_max_cluster_dia, parse_select_cond)
from clease.gui.constants import FOREGROUND_TEXT_COLOR
from clease.gui.constants import ECI_GRAPH_COLORS
from clease.gui.fitting_alg_editors import (FitAlgEditor, LassoEditor, L2Editor, BCSEditor,
                                            GAEditor)
from clease.gui.legend import Legend
from clease.gui import backup_folder

from threading import Thread
from kivy.uix.screenmanager import Screen
from kivy.utils import get_color_from_hex
from kivy_garden.graph import Graph, ScatterPlot, BarPlot, LinePlot
import numpy as np
import traceback


class ECIOptimizer(object):
    fit_page = None
    evaluator = None

    def optimize(self):
        try:
            self.fit_page.eci = self.evaluator.get_eci_dict()

            e_ce = self.evaluator.cf_matrix.dot(self.evaluator.eci)
            self.fit_page.e_ce = e_ce
            self.fit_page.e_dft = self.evaluator.e_dft
            self.fit_page.e_pred_leave_out = self.evaluator.e_pred_loo
            cv = self.evaluator.get_cv_score()
            rmse = self.evaluator.rmse() * 1000.0
            mae = self.evaluator.mae() * 1000.0
            self.fit_page.set_cv(cv)
            self.fit_page.set_rmse(rmse)
            self.fit_page.set_mae(mae)

            App.get_running_app().root.ids.status.text = 'Idle'
            self.fit_page.fitting_in_progress = False
            self.fit_page.update_plots()
        except Exception as exc:
            traceback.print_exc()
            App.get_running_app().root.ids.status.text = str(exc)
            self.fit_page.fitting_in_progress = False
            return


class GAClusterSelector(object):
    fit_page = None
    kwargs = None
    settings = None
    _pop_up = None

    def run(self):
        try:
            gen_without_change = self.kwargs.pop('gen_without_change')
            load_file = self.kwargs.pop('load_file')
            if not load_file:
                fname1 = backup_folder / 'ga_fit.csv'
                if fname1.exists():
                    fname1.unlink()

            self.kwargs['fname'] = str(backup_folder / 'ga_fit.csv')
            evaluator = Evaluate(self.settings)
            ga = GAFit(evaluator.cf_matrix, evaluator.e_dft, **self.kwargs)
            best_indx = ga.run(gen_without_change=gen_without_change)

            cf_names = [evaluator.cf_names[i] for i, flag in enumerate(best_indx) if flag]
            optimizer = ECIOptimizer()
            optimizer.fit_page = self.fit_page
            optimizer.evaluator = Evaluate(self.settings,
                                           fitting_scheme=LinearRegression(),
                                           cf_names=cf_names)
            optimizer.optimize()
        except Exception as exc:
            traceback.print_exc()
            App.get_running_app().root.ids.status.text = str(exc)


class FitPage(Screen):
    graphs_added = False
    # Initialize standard fitting params
    fitting_params = {'algorithm': 'LASSO', 'alpha': 0.001}
    eci = {}
    fit_result = {}
    _pop_up = None
    fitting_in_progress = False
    e_dft = None
    e_ce = None

    energy_plot = None
    eci_plots = []
    energy_graph = None
    eci_graph = None
    zero_line_energy = None

    legend = None

    def on_enter(self):
        self.load_fit_alg_settings(self.fitting_params['algorithm'])
        if self.legend is None:
            self.legend = Legend(self.ids.legend)
        if not self.graphs_added:
            self.energy_graph = Graph(xlabel='DFT energy (eV/atom)',
                                      ylabel="E_CE - E_DFT (meV/atom)",
                                      x_ticks_minor=0,
                                      x_ticks_major=10,
                                      y_ticks_major=10,
                                      y_grid_label=True,
                                      x_grid_label=True,
                                      padding=5,
                                      xlog=False,
                                      ylog=False,
                                      xmin=0.0,
                                      ymin=0.0,
                                      precision='%.2f')
            self.energy_plot = ScatterPlot(color=FOREGROUND_TEXT_COLOR, point_size=3)
            self.zero_line_energy = LinePlot(line_width=2, color=FOREGROUND_TEXT_COLOR)
            self.energy_graph.add_plot(self.energy_plot)
            self.energy_graph.add_plot(self.zero_line_energy)
            self.ids.energyPlot.add_widget(self.energy_graph)

            self.eci_graph = Graph(ylabel='ECI (eV/atom)',
                                   xlabel='Clusters',
                                   x_ticks_minor=0,
                                   x_ticks_major=10,
                                   y_ticks_major=10,
                                   y_grid_label=True,
                                   x_grid_label=True,
                                   padding=5,
                                   xlog=False,
                                   ylog=False,
                                   xmin=-1.0,
                                   ymin=0.0,
                                   precision='%.2f')

            for i in range(8):
                color = ECI_GRAPH_COLORS[i % len(ECI_GRAPH_COLORS)]
                color = get_color_from_hex(color)
                plot = BarPlot(color=color, bar_width=4, bar_spacing=1)
                self.eci_plots.append(plot)
                self.eci_graph.add_plot(plot)

            self.ids.eciPlot.add_widget(self.eci_graph)
            self.graphs_added = True

    def dismiss_popup(self):
        if self._pop_up is None:
            return

        if isinstance(self._pop_up.content, FitAlgEditor):
            self._pop_up.content.backup()
        self._pop_up.dismiss()
        self._pop_up = None

    def show_lasso_editor(self):
        content = LassoEditor(close=self.close_lasso_editor)
        self._pop_up = Popup(title="LASSO Editor",
                             content=content,
                             pos_hint={
                                 'right': 0.95,
                                 'top': 0.95
                             },
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def show_l2_editor(self):
        content = L2Editor(close=self.close_l2_editor)
        self._pop_up = Popup(title="L2 Editor",
                             content=content,
                             pos_hint={
                                 'right': 0.95,
                                 'top': 0.95
                             },
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def show_bcs_editor(self):
        content = BCSEditor(close=self.close_bcs_editor)

        self._pop_up = Popup(title="BCS Editor",
                             content=content,
                             pos_hint={
                                 'right': 0.95,
                                 'top': 0.95
                             },
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def show_ga_editor(self):
        content = GAEditor(close=self.close_ga_editor)

        self._pop_up = Popup(title="Genetic Algorithm Editor",
                             content=content,
                             pos_hint={
                                 'right': 0.95,
                                 'top': 0.95
                             },
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def close_l2_editor(self, alpha):
        self.fitting_params = {}
        self.fitting_params['algorithm'] = 'L2'
        self.fitting_params['alpha'] = float(alpha)
        self.dismiss_popup()

    def close_lasso_editor(self, alpha):
        self.fitting_params = {}
        self.fitting_params['algorithm'] = 'LASSO'
        self.fitting_params['alpha'] = float(alpha)
        self.dismiss_popup()

    def close_bcs_editor(self, shape_var, rate_var, shape_lamb, var_opt_start, init_lamb,
                         lamb_opt_start, max_iter, noise):
        self.fitting_params = {
            'algorithm': 'BCS',
            'shape_var': float(shape_var),
            'rate_var': float(rate_var),
            'shape_lamb': float(shape_lamb),
            'var_opt_start': float(var_opt_start),
            'init_lamb': float(init_lamb),
            'lamb_opt_start': float(lamb_opt_start),
            'max_iter': int(max_iter),
            'noise': float(noise)
        }
        self.dismiss_popup()

    def close_ga_editor(self, elitism, mut_prob, num_individuals, max_active, cost_func, load_file,
                        max_without_imp):
        if isinstance(load_file, str):
            load_file = load_file == 'True'
        self.fitting_params = {
            'algorithm': 'GA',
            'elitism': int(elitism),
            'mut_prob': float(mut_prob),
            'num_individuals': int(num_individuals),
            'max_active': int(max_active),
            'cost_func': cost_func,
            'load_file': load_file,
            'gen_without_change': int(max_without_imp)
        }
        self.dismiss_popup()

    def show_load_ECI_file_dialog(self):
        content = LoadDialog(load=self.load_eci_file, cancel=self.dismiss_popup)

        self._pop_up = Popup(title="Load ECI file",
                             content=content,
                             pos_hint={
                                 'right': 0.95,
                                 'top': 0.95
                             },
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def load_eci_file(self, path, filename):
        if len(filename) == 0:
            fname = path
        else:
            fname = filename[0]

        self.ids.eciFileInput.text = fname
        self.dismiss_popup()

    def launch_fit_alg_editor(self):
        if self.ids.fitAlgSpinner.text == 'LASSO':
            self.show_lasso_editor()
        elif self.ids.fitAlgSpinner.text == 'L2':
            self.show_l2_editor()
        elif self.ids.fitAlgSpinner.text == 'BCS':
            self.show_bcs_editor()
        elif self.ids.fitAlgSpinner.text == 'Genetic Algorithm':
            self.show_ga_editor()

    def save_eci(self):
        fname = self.ids.eciFileInput.text

        if not fname.endswith('.json'):
            msg = 'ECI file should be a JSON file'
            App.get_running_app().root.ids.status.text = msg

        try:
            with open(fname, 'w') as out:
                json.dump(self.eci, out, separators=(',', ': '), indent=2)
            msg = f"ECIs saved to {fname}"
            App.get_running_app().root.ids.status.text = msg
        except Exception as exc:
            traceback.print_exc()
            App.get_running_app().root.ids.status.text = str(exc)

    def fit_eci(self):
        if self.fitting_in_progress:
            # We already optimising ECIs
            return

        settings = App.get_running_app().root.settings

        if settings is None:
            msg = 'Settings not set. Call Update settings first.'
            App.get_running_app().root.ids.status.text = msg
            return

        if self.ids.maxClusterDiaCut.text == '':
            max_cluster_dia_cut = None
        else:
            try:
                max_cluster_dia_cut = \
                    parse_max_cluster_dia(self.ids.maxClusterDiaCut.text)
            except Exception as exc:
                traceback.print_exc()
                App.get_running_app().root.ids.status.text = str(exc)
                return

        if self.ids.maxClusterSizeCut.text == '':
            max_cluster_size_cut = None
        else:
            max_cluster_size_cut = int(self.ids.maxClusterSizeCut.text)

        k_fold = self.ids.kFoldInput.text
        if k_fold == '':
            msg = 'K-fold has to be given'
            App.get_running_app().root.ids.status.text = msg
            return
        k_fold = int(k_fold)

        num_rep = self.ids.numRepititionsInput.text
        if num_rep == '':
            num_rep = 1
        else:
            num_rep = int(num_rep)

        scheme = self.fitting_params.get('algorithm', None)
        if scheme is None:
            msg = 'Open the fitting scheme editor prior to the fit'
            App.get_running_app().root.ids.status.text = msg
            return

        select_cond = self.ids.dbSelectCondInput.text
        if select_cond == '':
            select_cond = None
        else:
            select_cond = parse_select_cond(select_cond)

        scheme = scheme.lower()
        alpha = 0.0
        if scheme in ['lasso', 'l2']:
            alpha = self.fitting_params['alpha']
        elif scheme == 'bcs':
            from clease import BayesianCompressiveSensing
            scheme = BayesianCompressiveSensing(
                shape_var=self.fitting_params['shape_var'],
                rate_var=self.fitting_params['rate_var'],
                shape_lamb=self.fitting_params['shape_lamb'],
                variance_opt_start=self.fitting_params['var_opt_start'],
                lamb_opt_start=self.fitting_params['lamb_opt_start'],
                maxiter=self.fitting_params['max_iter'],
                noise=self.fitting_params['noise'],
                init_lamb=self.fitting_params['init_lamb'])
        elif scheme == 'ga':
            ga_runner = GAClusterSelector()
            ga_runner.fit_page = self
            gen_without_change = self.fitting_params['gen_without_change']
            ga_runner.kwargs = {
                'mutation_prob': self.fitting_params['mut_prob'],
                'elitism': self.fitting_params['elitism'],
                'fname': None,
                'num_individuals': self.fitting_params['num_individuals'],
                'max_num_in_init_pool': self.fitting_params['max_active'],
                'cost_func': self.fitting_params['cost_func'].lower(),
                'gen_without_change': gen_without_change,
                'load_file': self.fitting_params['load_file']
            }
            ga_runner.settings = settings

            # As GA behaves a bit different from the other schemes, we have
            # a separate runner and return after the runner is finished.
            msg = 'Selecting clusters with GA..'
            App.get_running_app().root.ids.status.text = msg
            Thread(target=ga_runner.run).start()
            return

        if k_fold == -1:
            scoring_scheme = 'loocv'
        else:
            scoring_scheme = 'k-fold'

        try:
            evaluator = Evaluate(settings,
                                 fitting_scheme=scheme,
                                 alpha=alpha,
                                 max_cluster_size=max_cluster_size_cut,
                                 max_cluster_dia=max_cluster_dia_cut,
                                 nsplits=k_fold,
                                 num_repetitions=num_rep,
                                 scoring_scheme=scoring_scheme,
                                 select_cond=select_cond)

            App.get_running_app().root.ids.status.text = 'Optimizing ECIs...'
            eci_optimizer = ECIOptimizer()
            eci_optimizer.fit_page = self
            eci_optimizer.evaluator = evaluator
            self.fitting_in_progress = True

            Thread(target=eci_optimizer.optimize).start()
        except Exception as exc:
            traceback.print_exc()
            App.get_running_app().root.ids.status.text = str(exc)

    def set_cv(self, cv):
        self.ids.cvLabel.text = f"CV: {cv:.3f} meV/atom"

    def set_rmse(self, rmse):
        self.ids.rmseLabel.text = f"RMSE: {rmse:.3f} meV/atom"

    def set_mae(self, mae):
        self.ids.maeLabel.text = f"MAE: {mae:.3f} meV/atom"

    def _eci_has_been_fitted(self):
        return self.e_dft is not None and self.e_ce is not None

    def update_energy_plot(self, e_dft, e_ce):
        if not self._eci_has_been_fitted():
            msg = 'ECIs has not been fitted yet'
            App.get_running_app().root.ids.status.text = msg
            return

        xmin = np.min(e_dft)
        xmax = np.max(e_dft)
        ymin = np.min(e_ce - e_dft) * 1000.0
        ymax = np.max(e_ce - e_dft) * 1000.0

        x_rng = xmax - xmin
        xmin -= 0.05 * x_rng
        xmax += 0.05 * x_rng

        y_rng = ymax - ymin
        ymin -= 0.05 * y_rng
        ymax += 0.05 * y_rng

        self.energy_graph.xmin = float(xmin)
        self.energy_graph.xmax = float(xmax)
        self.energy_graph.ymin = float(ymin)
        self.energy_graph.ymax = float(ymax)
        self.energy_graph.x_ticks_major = float(xmax - xmin) / 10.0
        self.energy_graph.y_ticks_major = float(ymax - ymin) / 10.0
        self.energy_graph.y_grid_label = True
        self.energy_graph.x_grid_label = True
        self.energy_plot.points = [(y, (x - y) * 1000.0) for x, y in zip(e_ce, e_dft)]
        self.zero_line_energy.points = [(xmin, 0.0), (xmax, 0.0)]

    def _clear_eci_plots(self):
        for plot in self.eci_plots:
            plot.points = []

    def update_eci_plot(self, eci):
        eci_by_size = {}

        if len(eci) == 0:
            return

        max_size = 0
        for k, v in eci.items():
            size = int((k[1]))
            if size < 2:
                continue
            if size not in eci_by_size.keys():
                eci_by_size[size] = []
            eci_by_size[size].append(v)

            if size > max_size:
                max_size = size

        sorted_keys = sorted(list(eci_by_size.keys()))
        prev = 0

        xmax = len(list(eci.keys())) + len(sorted_keys)
        ymin = min([v for _, v in eci.items()])
        ymax = max([v for _, v in eci.items()])
        self._clear_eci_plots()
        for i, k in enumerate(sorted_keys):
            values = eci_by_size[k]
            indx = range(prev, prev + len(values))
            self.eci_plots[i].points = list(zip(indx, values))
            prev = indx[-1] + 2

        y_rng = ymax - ymin
        ymin -= 0.05 * y_rng
        ymax += 0.05 * y_rng

        self.eci_graph.xmax = int(xmax)
        self.eci_graph.ymin = float(ymin)
        self.eci_graph.ymax = float(ymax)
        self.eci_graph.x_ticks_major = float(xmax) / 10.0
        self.eci_graph.y_ticks_major = float(ymax - ymin) / 10.0

        num_rows = (max_size - 2) // 3 + 1
        num_cols = min([max_size - 1, 3])

        legend_items = []
        for k in sorted_keys:
            color = get_color_from_hex(ECI_GRAPH_COLORS[int(k) - 2])[:3]
            legend_items.append({'text': f"{k}-body", 'color': color})
        self.legend.setup(legend_items, num_rows=num_rows, num_cols=num_cols)

    def update_plots(self):
        if self.fitting_in_progress:
            msg = 'Performing ECI opmisation in the background. '
            msg += 'Wait until the process is finished.'
            App.get_running_app().root.ids.status.text = msg

        self.update_energy_plot(self.e_dft, self.e_ce)
        self.update_eci_plot(self.eci)

    def to_dict(self):
        return {
            'eci_file': self.ids.eciFileInput.text,
            'db_select_cond': self.ids.dbSelectCondInput.text,
            'max_cluster_size': self.ids.maxClusterSizeCut.text,
            'max_cluster_dia': self.ids.maxClusterDiaCut.text,
            'k_fold': self.ids.kFoldInput.text,
            'num_repetitions': self.ids.numRepititionsInput.text,
            'fit_alg': self.ids.fitAlgSpinner.text
        }

    def from_dict(self, data):
        self.ids.eciFileInput.text = data.get('eci_file', '')
        self.ids.dbSelectCondInput.text = data.get('db_select_cond', '')
        self.ids.maxClusterSizeCut.text = data.get('max_cluster_size', '')
        self.ids.maxClusterDiaCut.text = data.get('max_cluster_dia', '')
        self.ids.kFoldInput.text = data.get('k_fold', '10')
        self.ids.numRepititionsInput.text = data.get('num_repetitions', '1')
        self.ids.fitAlgSpinner.text = data.get('fit_alg', 'LASSO')

    def load_fit_alg_settings(self, text):
        fnames = {
            'LASSO': LassoEditor.backup_file,
            'L2': L2Editor.backup_file,
            'BCS': BCSEditor.backup_file,
            'Genetic Algorithm': GAEditor.backup_file
        }

        closing_methods = {
            'LASSO': self.close_lasso_editor,
            'L2': self.close_l2_editor,
            'BCS': self.close_bcs_editor,
            'Genetic Algorithm': self.close_ga_editor
        }

        editors = {
            'LASSO': LassoEditor(),
            'L2': L2Editor(),
            'BCS': BCSEditor(),
            'Genetic Algorithm': GAEditor()
        }
        fname = fnames.get(text, None)
        close = closing_methods[text]

        app = App.get_running_app()
        if fname is None:
            app.root.ids.status.text = 'Unkown fitting scheme'
            return

        full_name = backup_folder / fname
        if not full_name.exists():
            # Create the file
            editors[text].backup()

        args = []
        with full_name.open('r') as fd:
            for line in fd:
                args.append(line.strip())
        args = args[::-1]
        try:
            close(*args)
            msg = 'Fit settings set the ones used '
            msg += 'last time'
            app.root.ids.status.text = msg
        except Exception:
            msg = 'Failed load previous fitting settings. '
            msg += 'Please set your settings again in the editor.'
            app.root.ids.status.text = msg



"""Module for setting up pseudospins and basis functions."""
from abc import ABC, abstractmethod
import math
from typing import List, Dict, Optional, Sequence
import numpy as np
from clease.gramSchmidthMonomials import GramSchmidtMonimial

__all__ = ('BasisFunction', 'Polynomial', 'Trigonometric', 'BinaryLinear')


class BasisFunction(ABC):
    """Base class for all Basis Functions."""

    def __init__(self, unique_elements: Sequence[str]) -> None:
        self.name = "generic"
        self.unique_elements = unique_elements
        if self.num_unique_elements < 2:
            raise ValueError("Systems must have more than 1 type of element.")

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BasisFunction):
            return False
        return self.name == other.name and \
            self.unique_elements == other.unique_elements

    @property
    def unique_elements(self) -> List[str]:
        return self._unique_elements

    @unique_elements.setter
    def unique_elements(self, elements):
        self._unique_elements = sorted(set(elements))

    @property
    def num_unique_elements(self) -> int:
        return len(self.unique_elements)

    @property
    def spin_dict(self) -> Dict[str, int]:
        return self.get_spin_dict()

    @property
    def basis_functions(self) -> List[Dict[str, float]]:
        return self.get_basis_functions()

    @abstractmethod
    def get_spin_dict(self):
        """Get spin dictionary."""

    @abstractmethod
    def get_basis_functions(self):
        """Get basis function."""

    # pylint: disable=no-self-use
    def customize_full_cluster_name(self, full_cluster_name: str) -> str:
        """Customize the full cluster names. Default is to do nothing."""
        return full_cluster_name

    def todict(self) -> dict:
        """
        Create a dictionary representation of the basis function class
        """
        return {'name': self.name, 'unique_elements': self.unique_elements}


class Polynomial(BasisFunction):
    """Pseudospin and basis function from Sanchez et al.

    Sanchez, J. M., Ducastelle, F. and Gratias, D. (1984).
    Generalized cluster description of multicomponent systems.
    Physica A: Statistical Mechanics and Its Applications, 128(1-2), 334-350.
    """

    def __init__(self, unique_elements: List[str]):
        BasisFunction.__init__(self, unique_elements)
        self.name = "polynomial"

    def get_spin_dict(self) -> Dict[str, int]:
        """Define pseudospins for all consistuting elements."""
        gram_schmidt = GramSchmidtMonimial(self.num_unique_elements)
        spin_values = gram_schmidt.values
        spin_dict = {}
        for x in range(self.num_unique_elements):
            spin_dict[self.unique_elements[x]] = spin_values[x]
        return spin_dict

    def get_basis_functions(self) -> List[Dict[str, float]]:
        """Create basis functions to guarantee the orthonormality."""
        gram_schmidt = GramSchmidtMonimial(self.num_unique_elements)
        gram_schmidt.build()
        return gram_schmidt.basis_functions(self.unique_elements)


class Trigonometric(BasisFunction):
    """Pseudospin and basis function from van de Walle.

    van de Walle, A. (2009).
    Multicomponent multisublattice alloys, nonconfigurational entropy and other
    additions to the Alloy Theoretic Automated Toolkit. Calphad, 33(2),
    266-278.
    """

    def __init__(self, unique_elements: List[str]):
        BasisFunction.__init__(self, unique_elements)
        self.name = "trigonometric"

    def get_spin_dict(self) -> Dict[str, int]:
        """Define pseudospins for all consistuting elements."""
        spin_values = list(range(self.num_unique_elements))
        spin_dict = {}
        for x in range(self.num_unique_elements):
            spin_dict[self.unique_elements[x]] = spin_values[x]
        return spin_dict

    def get_basis_functions(self) -> List[Dict[str, float]]:
        """Create basis functions to guarantee the orthonormality."""
        alpha = list(range(1, self.num_unique_elements))
        bf_list = []

        for a in alpha:
            bf = {}
            for key, value in self.spin_dict.items():
                var = 2 * np.pi * math.ceil(a / 2.) * value
                var /= self.num_unique_elements
                if a % 2 == 1:
                    bf[key] = -np.cos(var) + 0.
                else:
                    bf[key] = -np.sin(var) + 0.

            # normalize the basis function
            sum_ = sum(bf[key] * bf[key] for key in self.spin_dict)
            normalization_factor = np.sqrt(self.num_unique_elements / sum_)

            for key, value in bf.items():
                bf[key] = value * normalization_factor

            bf_list.append(bf)

        return bf_list


def _kronecker(i: int, j: int) -> int:
    """Kronecker delta function."""
    if i == j:
        return 1
    return 0


class BinaryLinear(BasisFunction):
    """Pseudospin and basis function from Zhang and Sluiter.

    Zhang, X. and Sluiter M.
    Cluster expansions for thermodynamics and kinetics of multicomponent
    alloys.
    Journal of Phase Equilibria and Diffusion 37(1) 44-52.
    """

    def __init__(self, unique_elements: List[str], redundant_element: Optional[str] = "auto"):
        if redundant_element == "auto":
            self.redundant_element = sorted(unique_elements)[0]
        else:
            self.redundant_element = redundant_element
        BasisFunction.__init__(self, unique_elements)
        self.name = "binary_linear"

    def get_spin_dict(self) -> Dict[str, int]:
        """Define pseudospins for all consistuting elements."""
        spin_values = list(range(self.num_unique_elements))
        spin_dict = {}
        for x in range(self.num_unique_elements):
            spin_dict[self.unique_elements[x]] = spin_values[x]
        return spin_dict

    def get_basis_functions(self) -> List[Dict[str, float]]:
        """Create orthonormal basis functions.

        Due to the constraint that any site is occupied by exactly one element,
        we only need to track N-1 species if there are N species.
        Hence, the first element specified is redundant, and will not
        have a basis function.
        """
        bf_list = []
        num_bf = self.num_unique_elements
        for bf_num in range(num_bf):
            if self.unique_elements[bf_num] == self.redundant_element:
                continue
            new_bf = {
                symb: float(_kronecker(i, bf_num)) for i, symb in enumerate(self.unique_elements)
            }
            bf_list.append(new_bf)
        return bf_list

    def _decoration2element(self, dec_num: int) -> str:
        """Get the element with its basis function equal to 1."""
        bf = self.basis_functions[dec_num]
        for k, v in bf.items():
            if v == 1:
                return k
        raise ValueError("Did not find any element where the value is 1.")

    def customize_full_cluster_name(self, full_cluster_name: str) -> str:
        """Translate the decoration number to element names."""
        dec = full_cluster_name.rsplit("_", 1)[1]
        name = full_cluster_name.rsplit("_", 1)[0]
        new_dec = ""
        for decnum in dec:
            element = self._decoration2element(int(decnum))
            new_dec += f"{element}"
        return name + "_" + new_dec

    def todict(self) -> dict:
        """
        Creates a dictionary representation of the class
        """
        dct_rep = BasisFunction.todict(self)
        dct_rep['redundant_element'] = self.redundant_element
        return dct_rep



from kivy.uix.screenmanager import Screen
from kivy_garden.graph import Graph, LinePlot
from clease.gui.constants import (FOREGROUND_TEXT_COLOR, META_DYN_MSG)
from kivy.app import App
from kivy.uix.popup import Popup
from clease.gui.help_message_popup import HelpMessagePopup
from clease.gui.conc_obs_editor import ConcObsEditor
from clease.gui.sgc_editor import SGCEditor
from threading import Thread
from clease.gui.meta_dyn_runner import MetaDynRunner
from clease.montecarlo import BinnedBiasPotential
from clease.montecarlo.observers import ConcentrationObserver
import json
import os
import numpy as np
from clease.tools import species_chempot2eci


class MetaDynPage(Screen):
    _pop_up = None
    observer_params = {}
    ensemble_params = {}
    visit_graph = None
    visit_plot = None
    visit_mean_plot = None
    visit_conv_plot = None
    beta_free_energy_graph = None
    beta_free_energy_plot = None
    mc_is_running = False
    _meta_dyn_sampler = None

    def on_enter(self):
        app = App.get_running_app()
        app.root.mc_type_screen = 'MetaDynPage'

        try:
            if self.observer_params == {}:
                # Open and close the observer editor
                self.launch_observer_editor(open_popup=False)
                self._pop_up.content.ids.closeButton.dispatch('on_release')
            if self.ensemble_params == {}:
                self.launch_ensemble_editor(open_popup=False)
                self._pop_up.content.ids.closeButton.dispatch('on_release')
        except Exception:
            pass

        if self.beta_free_energy_graph is None:
            self.beta_free_energy_graph = Graph(xlabel='Collective variable',
                                                ylabel="G/kT",
                                                x_ticks_minor=0,
                                                x_ticks_major=10,
                                                y_ticks_major=10,
                                                y_grid_label=True,
                                                x_grid_label=True,
                                                padding=5,
                                                xlog=False,
                                                ylog=False,
                                                xmin=0.0,
                                                ymin=0.0,
                                                precision='%1.1e')

            self.beta_free_energy_plot = LinePlot(line_width=1, color=FOREGROUND_TEXT_COLOR)
            self.beta_free_energy_graph.add_plot(self.beta_free_energy_plot)
            self.ids.freeEnergyPlot.add_widget(self.beta_free_energy_graph)

            self.visit_graph = Graph(xlabel='Collective variable',
                                     ylabel="Num. visits",
                                     x_ticks_minor=0,
                                     x_ticks_major=10,
                                     y_ticks_major=10,
                                     y_grid_label=True,
                                     x_grid_label=True,
                                     padding=5,
                                     xlog=False,
                                     ylog=False,
                                     xmin=0.0,
                                     ymin=0.0,
                                     precision='%1.1e')
            self.visit_plot = LinePlot(line_width=1, color=FOREGROUND_TEXT_COLOR)
            self.visit_mean_plot = LinePlot(line_width=1, color=FOREGROUND_TEXT_COLOR)
            self.visit_conv_plot = LinePlot(line_width=1, color=FOREGROUND_TEXT_COLOR)
            self.visit_graph.add_plot(self.visit_plot)
            self.visit_graph.add_plot(self.visit_mean_plot)
            self.visit_graph.add_plot(self.visit_conv_plot)
            self.ids.visitHistogram.add_widget(self.visit_graph)

    def bind_meta_dyn_sampler(self, sampler):
        self._meta_dyn_sampler = sampler

    def detach_meta_dyn_sampler(self):
        self._meta_dyn_sampler = None

    def abort_mc(self):
        if self._meta_dyn_sampler is not None:
            self._meta_dyn_sampler.quit = True
            app = App.get_running_app()
            app.root.ids.status.text = META_DYN_MSG['abort_mc']

    def dismiss_popup(self):
        if self._pop_up is None:
            return
        self._pop_up.dismiss()
        self._pop_up = None

    def show_temperature_help(self):
        msg = 'Temperature used during sampling given in Kelvin'
        content = HelpMessagePopup(close=self.dismiss_popup, msg=msg)
        self._pop_up = Popup(title="Temperature input",
                             content=content,
                             pos_hint={
                                 'right': 0.95,
                                 'top': 0.95
                             },
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def show_var_help(self):
        msg = 'Collective variable used to map out the \n'
        msg += 'free energy.'
        content = HelpMessagePopup(close=self.dismiss_popup, msg=msg)
        self._pop_up = Popup(title="Collective variable",
                             content=content,
                             pos_hint={
                                 'right': 0.95,
                                 'top': 0.95
                             },
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def show_range_help(self):
        msg = 'The simulation will be limited to configurations\n'
        msg += 'where the collective variable is within the \n'
        msg += 'specified range.'
        content = HelpMessagePopup(close=self.dismiss_popup, msg=msg)
        self._pop_up = Popup(title="Range help",
                             content=content,
                             pos_hint={
                                 'right': 0.95,
                                 'top': 0.95
                             },
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def show_flatness_help(self):
        msg = 'The histogram of visits is considered to be\n'
        msg += 'flat, when the bin with the minimum value\n'
        msg += 'is larger that <flatness criteria> times the\n'
        msg += 'mean.\n'
        content = HelpMessagePopup(close=self.dismiss_popup, msg=msg)
        self._pop_up = Popup(title="Flatness help",
                             content=content,
                             pos_hint={
                                 'right': 0.95,
                                 'top': 0.95
                             },
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def show_nbins_help(self):
        msg = 'Number of intervals which is used to partition\n'
        msg += 'the interval.\n'
        content = HelpMessagePopup(close=self.dismiss_popup, msg=msg)
        self._pop_up = Popup(title="Num. bins help",
                             content=content,
                             pos_hint={
                                 'right': 0.95,
                                 'top': 0.95
                             },
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def show_modfactor_help(self):
        msg = 'The modification factor is a penalisation value \n'
        msg += 'added to the free energy curve when the sampler\n'
        msg += 'visits a particular bin. It is given in units of\n'
        msg += 'kT. Hence, if it is 0.01, it means that the energy\n'
        msg += 'is penalised by 0.01*kT.'
        content = HelpMessagePopup(close=self.dismiss_popup, msg=msg)
        self._pop_up = Popup(title="Mod. factor help",
                             content=content,
                             pos_hint={
                                 'right': 0.95,
                                 'top': 0.95
                             },
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def show_max_sweep_help(self):
        msg = 'Maximum number of sweeps before the calculation\n'
        msg += 'will be terminated. If it converges before reaching\n'
        msg += 'this number, it stops earlier\n'
        content = HelpMessagePopup(close=self.dismiss_popup, msg=msg)
        self._pop_up = Popup(title="Max. sweeps help",
                             content=content,
                             pos_hint={
                                 'right': 0.95,
                                 'top': 0.95
                             },
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def close_conc_editor(self, element):
        self._pop_up.content.backup()
        self.observer_params = {'name': 'Concentration', 'element': element}
        self.dismiss_popup()

    def launch_observer_editor(self, var=None, open_popup=True):
        if var is None:
            var = self.ids.varSpinner.text

        content = None
        if var == 'Concentration':
            default_elem = sorted(self.unique_symbols())[0]
            content = ConcObsEditor(close=self.close_conc_editor, default_element=default_elem)

        if content is None:
            return
        self._pop_up = Popup(title=var,
                             content=content,
                             pos_hint={
                                 'right': 0.95,
                                 'top': 0.95
                             },
                             size_hint=(0.9, 0.9))
        if open_popup:
            self._pop_up.open()

    def launch_ensemble_editor(self, text=None, open_popup=True):
        if text is None:
            text = self.ids.ensembleSpinner.text

        content = None
        if text == 'Semi-grand canonical':
            symbols = self.unique_symbols()
            default_symbs = ', '.join(symbols)
            def_chem_pot = ', '.join([f"{s}: 0.0" for s in symbols[:-1]])
            content = SGCEditor(close=self.close_sgc_editor,
                                symbols=default_symbs,
                                chem_pot=def_chem_pot)

        if content is None:
            return

        self._pop_up = Popup(title=text,
                             content=content,
                             pos_hint={
                                 'right': 0.95,
                                 'top': 0.95
                             },
                             size_hint=(0.9, 0.9))
        if open_popup:
            self._pop_up.open()

    def close_sgc_editor(self, symbols, chem_pot):
        self._pop_up.content.backup()
        self.ensemble_params = {
            'name': 'Semi-grand canonical',
            'symbols': symbols,
            'chem_pot': chem_pot
        }
        self.dismiss_popup()

    def run(self):
        """
        Run the meta dynamics sampler
        """
        if self.mc_is_running:
            return
        app = App.get_running_app()
        settings = app.root.settings
        if settings is None:
            app.root.ids.status.text = META_DYN_MSG['settings_is_none']
            return

        if len(settings.concentration.basis_elements) > 1:
            app.root.ids.status.text = META_DYN_MSG['more_than_one_basis']
            return

        # Load ECI
        mc_header = app.root.ids.sm.get_screen('MCHeader')
        main_mc_page = mc_header.ids.sm.get_screen('MCMainPage')
        eci_file = main_mc_page.ids.eciFileInput.text

        if not os.path.exists(eci_file):
            msg = META_DYN_MSG['no_eci'] + f"{eci_file}"
            app.root.ids.status.text = msg
            return

        with open(eci_file, 'r') as infile:
            eci = json.load(infile)

        try:
            size = int(main_mc_page.ids.sizeInput.text)
        except Exception as exc:
            app.root.ids.status.text = str(exc)
            return

        if app.root.active_template_is_mc_cell:
            app.root.ids.status.text = META_DYN_MSG['mc_cell_is_template']
            return

        atoms = settings.atoms * (size, size, size)

        ensemble = self.ids.ensembleSpinner.text

        try:
            T = float(self.ids.tempInput.text)
            xmin = float(self.ids.varMinInput.text)
            xmax = float(self.ids.varMaxInput.text)
            nbins = int(self.ids.nbinsInput.text)
            max_sweeps = int(self.ids.maxSweepsInput.text)
            flat_limit = float(self.ids.flatInput.text)
            mod_factor = float(self.ids.modInput.text)
            update_int = int(self.ids.plotIntInput.text)
        except Exception as exc:
            app.root.ids.status.text = str(exc)
            return

        getter = None
        variable = self.ids.varSpinner.text
        if variable == 'Concentration':
            if 'element' not in self.observer_params.keys():
                msg = META_DYN_MSG['var_editor_not_launched']
                app.root.ids.status.text = msg
                return
            elem = self.observer_params['element']
            getter = ConcentrationObserver(atoms, element=elem)
        else:
            msg = META_DYN_MSG['unkown_var'] + variable
            app.root.ids.status.text = msg
            return

        pot = BinnedBiasPotential(xmin=xmin, xmax=xmax, nbins=nbins, getter=getter)

        fname = self.ids.backupInput.text + '.json'
        if os.path.exists(fname):
            with open(fname, 'r') as infile:
                data = json.load(infile)
            pot.from_dict(data['bias_pot'])

        if ensemble == 'Semi-grand canonical':
            if 'symbols' not in self.ensemble_params.keys():
                app.root.ids.status.text = META_DYN_MSG['launch_ens_editor']
                return
            chem_pot = self.ensemble_params['chem_pot']
            bf_list = settings.basis_functions
            chem_pot_eci = species_chempot2eci(bf_list, chem_pot)
            for k, v in chem_pot_eci.items():
                eci[k] = eci.get(k, 0.0) - v
        else:
            app.root.ids.status.text = META_DYN_MSG['unknown_ens'] + ensemble
            return

        runner = MetaDynRunner(atoms=atoms,
                               meta_page=self,
                               max_sweeps=max_sweeps,
                               settings=settings,
                               app_root=app.root,
                               eci=eci,
                               status=app.root.ids.status,
                               mc_params=self.ensemble_params,
                               T=T,
                               bias=pot,
                               flat=flat_limit,
                               mod_factor=mod_factor,
                               backup_file=fname,
                               update_interval=update_int)
        Thread(target=runner.run).start()

    def to_dict(self):
        return {
            'temperature': self.ids.tempInput.text,
            'collective_var': self.ids.varSpinner.text,
            'min_collective_var': self.ids.varMinInput.text,
            'max_collective_var': self.ids.varMaxInput.text,
            'flatness': self.ids.flatInput.text,
            'nbins': self.ids.nbinsInput.text,
            'backup': self.ids.backupInput.text,
            'ensemble': self.ids.ensembleSpinner.text,
            'mod_factor': self.ids.modInput.text,
            'max_sweeps': self.ids.maxSweepsInput.text,
            'plot_update_interval': self.ids.plotIntInput.text
        }

    def from_dict(self, dct):
        """
        Initialise the fields from a dictionary
        """
        self.ids.tempInput.text = dct.get('temperature', '500')
        self.ids.varSpinner.text = dct.get('collective_var', 'Concentration')
        self.ids.varMinInput.text = dct.get('min_collective_var', '0.0')
        self.ids.varMaxInput.text = dct.get('max_collective_var', '1.0')
        self.ids.flatInput.text = dct.get('flatness', '0.8')
        self.ids.nbinsInput.text = dct.get('nbins', '50')
        self.ids.backupInput.text = dct.get('backup', 'metadyn')
        self.ids.ensembleSpinner.text = dct.get('ensemble', 'Semi-grand canonical')
        self.ids.modInput.text = dct.get('mod_factor', '0.1')
        self.ids.maxSweepsInput.text = dct.get('max_sweep', '10000')
        self.ids.plotIntInput.text = dct.get('plot_update_interval', '10')

    def update_free_energy_plot(self, x, betaG):
        """
        Updates the free energy plot

        Parameters:

        x: array
            Collective variable

        betaG: array
            Array with the free energy (divided by kT)
        """
        xmin = np.min(x)
        xmax = np.max(x)
        x_rng = xmax - xmin
        xmin -= 0.05 * x_rng
        xmax += 0.05 * x_rng

        betaG -= betaG[0]
        ymin = np.min(betaG)
        ymax = np.max(betaG)
        y_rng = ymax - ymin
        ymin -= 0.05 * y_rng
        ymax += 0.05 * y_rng

        self.beta_free_energy_graph.xmin = float(xmin)
        self.beta_free_energy_graph.xmax = float(xmax)
        self.beta_free_energy_graph.ymin = float(ymin)
        self.beta_free_energy_graph.ymax = float(ymax)
        self.beta_free_energy_graph.x_ticks_major = float(xmax - xmin) / 10.0
        self.beta_free_energy_graph.y_ticks_major = float(ymax - ymin) / 10.0
        self.beta_free_energy_graph.y_grid_label = True
        self.beta_free_energy_graph.x_grid_label = True
        self.beta_free_energy_plot.points = list(zip(x, betaG))

    def update_visit_plot(self, x, visits, flat_criteria):
        """
        Update the visit graph

        Parameters:

        x: array
            Array with the collective variables

        visits: array
            Array with the visit histogram

        flat_criteria: float
            Criteria used to judge when a histogram is flat
        """
        xmin = np.min(x)
        xmax = np.max(x)
        x_rng = xmax - xmin
        xmin -= 0.05 * x_rng
        xmax += 0.05 * x_rng

        ymin = np.min(visits)
        ymax = np.max(visits)
        y_rng = ymax - ymin
        ymin -= 0.05 * y_rng
        ymax += 0.05 * y_rng

        self.visit_graph.xmin = float(xmin)
        self.visit_graph.xmax = float(xmax)
        self.visit_graph.ymin = float(ymin)
        self.visit_graph.ymax = float(ymax)
        self.visit_graph.x_ticks_major = float(xmax - xmin) / 10.0
        self.visit_graph.y_ticks_major = float(ymax - ymin) / 10.0
        self.visit_graph.y_grid_label = True
        self.visit_graph.x_grid_label = True
        self.visit_plot.points = list(zip(x, visits))
        mean = np.mean(visits)
        self.visit_mean_plot.points = [(xmin, mean), (xmax, mean)]
        conv = flat_criteria * mean
        self.visit_conv_plot.points = [(xmin, conv), (xmax, conv)]

    def clear_backup(self):
        fname = self.ids.backupInput.text + '.json'

        if os.path.exists(fname):
            os.remove(fname)

    def unique_symbols(self):
        app = App.get_running_app()
        settings = app.root.settings

        if settings is None:
            app.root.ids.status.text = 'Settings is not set'
            return ['']

        unique_symb = settings.unique_element_without_background()
        return unique_symb



from typing import Sequence, List, Set
from abc import abstractmethod, ABC
from ase import Atoms
import numpy as np
from clease.montecarlo.constraints import MCConstraint
from clease.tools import SystemChange
from clease.montecarlo.swap_move_index_tracker import SwapMoveIndexTracker

__all__ = ('TrialMoveGenerator', 'RandomFlip', 'RandomSwap')

DEFAULT_MAX_ATTEMPTS = 10000


class TrialMoveGenerator(ABC):
    """
    Class for producing trial moves.

    :param max_attempts: Maximum number of attempts to try to find a move that passes
        the constraints. If not constraints are added, this has no effect.
    """

    def __init__(self, max_attempts: int = DEFAULT_MAX_ATTEMPTS):
        super().__init__()
        self._constraints = []
        self.max_attempts = max_attempts

    def initialize(self, atoms: Atoms) -> None:
        """
        Initialize the generator.

        :param atoms: Atoms object used in the simulation
        """

    def remove_constraints(self) -> None:
        """
        Remove all constraints
        """
        self._constraints = []

    @property
    def name(self):
        return type(self).__name__

    def add_constraint(self, cnst: MCConstraint):
        """
        Add a constraint to the generator

        :param cnst: Constraint that must be satisfied for all trial moves
        """
        self._constraints.append(cnst)

    def _change_ok(self, system_changes: Sequence[SystemChange]) -> bool:
        for cnst in self._constraints:
            if not cnst(system_changes):
                return False
        return True

    @abstractmethod
    def get_single_trial_move(self) -> Sequence[SystemChange]:
        """
        Return a single trial move, must be implemented in sub-classes
        """

    def get_trial_move(self) -> Sequence[SystemChange]:
        """
        Produce a trial move that is consistent with all cosntraints
        """
        for _ in range(self.max_attempts):
            move = self.get_single_trial_move()
            if self._change_ok(move):
                return move
        raise RuntimeError("Can not produce a valid trial move")

    def on_move_accepted(self, change: SystemChange) -> None:
        """
        Callback that is called by Monte Carlo after each accepted move

        :param change: Trial move
        """

    def on_move_rejected(self, change: SystemChange) -> None:
        """
        Callback that is called after a move is rejected

        :param change: Trial move
        """


class RandomFlip(TrialMoveGenerator):
    """
    Generate trial moves where the symbol at a given site is flipped

    :param symbols: Set with all symbols considered in a move
    :param atoms: Atoms object for the simulation
    :param indices: List with all indices that should be considered. If None, all indices are
        considered
    """

    def __init__(self, symbols: Set[str], atoms: Atoms, indices: List[int] = None, **kwargs):
        super().__init__(**kwargs)
        self.symbols = symbols
        self.atoms = atoms

        if indices is None:
            self.indices = list(range(len(self.atoms)))
        else:
            self.indices = indices

    def get_single_trial_move(self) -> List[SystemChange]:
        pos = np.random.choice(self.indices)
        old_symb = self.atoms[pos].symbol
        new_symb = np.random.choice([s for s in self.symbols if s != old_symb])
        return [SystemChange(index=pos, old_symb=old_symb, new_symb=new_symb)]


class RandomSwap(TrialMoveGenerator):
    """
    Produce random swaps

    :param atoms: Atoms object in the MC simulation
    :param indices: List with indices that can be chosen from. If None, all indices
        can be chosen.
    """

    def __init__(self, atoms: Atoms, indices: List[int] = None, **kwargs):
        super().__init__(**kwargs)
        self.tracker = SwapMoveIndexTracker()
        self.indices = indices
        self.initialize(atoms)

    def initialize(self, atoms: Atoms) -> None:
        """
        Initializes the atoms tracker
        """
        self.tracker.init_tracker(atoms)
        if self.indices:
            self.tracker.filter_indices(self.indices)

    def get_single_trial_move(self) -> List[SystemChange]:
        """
        Create a swap move
        """
        symb_a = np.random.choice(self.tracker.symbols)
        symb_b = np.random.choice([s for s in self.tracker.symbols if s != symb_a])
        rand_pos_a = self.tracker.get_random_indx_of_symbol(symb_a)
        rand_pos_b = self.tracker.get_random_indx_of_symbol(symb_b)
        return [
            SystemChange(index=rand_pos_a, old_symb=symb_a, new_symb=symb_b),
            SystemChange(index=rand_pos_b, old_symb=symb_b, new_symb=symb_a)
        ]

    def on_move_accepted(self, change: Sequence[SystemChange]):
        self.tracker.update_swap_move(change)



"""Module for calculating correlation functions."""
import logging
from ase.atoms import Atoms
from ase.db import connect

from clease.settings import ClusterExpansionSettings
from clease.tools import wrap_and_sort_by_position
from clease_cxx import PyCEUpdater

logger = logging.getLogger(__name__)
__all__ = ('CorrFunction', 'ClusterNotTrackedError')


class ClusterNotTrackedError(Exception):
    pass


class CorrFunction:
    """Calculate the correlation function.

    Parameters:

    settings: settings object

    parallel: bool (optional)
        specify whether or not to use the parallel processing for `get_cf`
        method.

    num_core: int or "all" (optional)
        specify the number of cores to use for parallelization.
    """

    def __init__(self, settings, parallel=False, num_core="all"):
        if not isinstance(settings, ClusterExpansionSettings):
            raise TypeError("setting must be CEBulk or CECrystal " "object")
        self.settings = settings

    def get_cf(self, atoms):
        """
        Calculate correlation functions for all possible clusters and return
        them in a dictionary format.

        Parameters:

        atoms: Atoms object
        """
        if not isinstance(atoms, Atoms):
            raise TypeError('atoms must be an Atoms object')
        cf_names = self.settings.all_cf_names
        return self.get_cf_by_names(atoms, cf_names)

    def get_cf_by_names(self, atoms, cf_names, warm=False):
        """
        Calculate correlation functions of the specified clusters and return
        them in a dictionary format.

        Parameters:

        atoms: Atoms object

        cf_names: list
            names of correlation functions that will be calculated for
            the structure provided in atoms
        """

        if isinstance(atoms, Atoms):
            self.check_cell_size(atoms)
        else:
            raise TypeError('atoms must be Atoms object')

        self._confirm_cf_names_exists(cf_names)

        eci = {name: 1.0 for name in cf_names}
        cf = {name: 1.0 for name in cf_names}
        updater = PyCEUpdater(atoms, self.settings, cf, eci, self.settings.cluster_list)
        cf = updater.calculate_cf_from_scratch(atoms, cf_names)
        return cf

    def reconfigure_db_entries(self, select_cond=None, verbose=False):
        """Reconfigure the correlation function values of the entries in DB.

        Parameters:

        select_cond: list
            -None (default): select every item in DB with
                             "struct_type='initial'"
            -else: select based on the condictions provided
                  (struct_type='initial' is not automatically included)

        verbose: bool
            print the progress of reconfiguration if set to *True*
        """
        tab_name = "{}_cf".format(self.settings.basis_func_type.name)
        db = connect(self.settings.db_name)
        db.delete_external_table(tab_name)
        select = []
        if select_cond is not None:
            for cond in select_cond:
                select.append(cond)
        else:
            select = [('struct_type', '=', 'initial')]

        # get how many entries need to be reconfigured
        row_ids = [row.id for row in db.select(select)]
        num_reconf = len(row_ids)
        msg = f"{num_reconf} entries will be reconfigured"
        logger.info(msg)
        if verbose:
            print(msg)
        for count, row_id in enumerate(row_ids):

            msg = f"updating {count+1} of {num_reconf} entries"
            if verbose:
                print(msg, end="\r")
            logger.debug(msg)
            atoms = wrap_and_sort_by_position(db.get(id=row_id).toatoms())
            cf = self.get_cf(atoms)
            db.update(row_id, external_tables={tab_name: cf})

        if verbose:
            print("\nreconfiguration completed")
        logger.info('Reconfiguration complete')

    def reconfigure_inconsistent_cf_table_entries(self):
        """Find and correct inconsistent correlation functions in table."""
        inconsistent_ids = self.check_consistency_of_cf_table_entries()

        if len(inconsistent_ids) == 0:
            return True

        logger.info('Reconfiguring correlation functions')
        for count, id in enumerate(inconsistent_ids):
            logger.debug("Updating %s of %s entries (id %s)", count + 1, len(inconsistent_ids), id)
            self.reconfigure_db_entries(select_cond=[('id', '=', id)], verbose=False)
        logger.info("Reconfiguration completed")

    def check_consistency_of_cf_table_entries(self):
        """Get IDs of the structures with inconsistent correlation functions.

        Note: consisent structures have the exactly the same list of cluster
              names as stored in settings.cf_names.
        """
        db = connect(self.settings.db_name)
        tab_name = f"{self.settings.basis_func_type.name}_cf"
        cf_names = sorted(self.settings.all_cf_names)
        inconsistent_ids = []
        for row in db.select('struct_type=initial'):
            tab_entries = row.get(tab_name, {})
            row_cnames = sorted(list(tab_entries.keys()))
            if row_cnames != cf_names:
                inconsistent_ids.append(row.id)

        if len(inconsistent_ids) > 0:
            logger.warning("%d inconsistent entries found in table %s", len(inconsistent_ids),
                           tab_name)
            for id in inconsistent_ids:
                logger.warning("  id: %s, name: %s", id, db.get(id).name)
        else:
            logger.info("'%s' table has no inconsistent entries.", tab_name)
        return inconsistent_ids

    def check_cell_size(self, atoms):
        """Check the size of provided cell and create a template if necessary.

        Parameters:

        atoms: Atoms object
            Unrelaxed structure
        """
        self.settings.set_active_template(atoms=atoms)
        return atoms

    def _cf_name_exists(self, cf_name):
        """Return True if cluster name exists. Otherwise False.

        Parameters:

        cluster_name: str
            Cluster name to check
        """
        return cf_name in self.settings.all_cf_names

    def _confirm_cf_names_exists(self, cf_names):
        if not set(cf_names).issubset(self.settings.all_cf_names):
            raise ClusterNotTrackedError("The correlation function of non-existing cluster is "
                                         "requested, but the name does not exist in "
                                         "ClusterExpansionSettings. Check that the cutoffs are "
                                         "correct, and try to run reconfigure_settings")



from typing import Dict, Sequence
from abc import ABCMeta
from ase import Atoms
from clease.tools import SystemChange

__all__ = ('BarrierModel', 'SSTEBarrier')


# pylint: disable=too-few-public-methods
class BarrierModel(metaclass=ABCMeta):
    """
    Base barrier model class
    """

    def __call__(self, atoms: Atoms, system_change: Sequence[SystemChange]) -> float:
        raise NotImplementedError("__call__ must be implemented in derived classes")


# pylint: disable=too-few-public-methods
class SSTEBarrier(BarrierModel):
    """
    SSTE is short for Solid Solution Total Energy. It is a simple barrier
    model where the barrier is modelled by the dilute barrier, corrected by
    the change in total energy.

    E_a = Q_s + (E_2 - E_1)/2

    where E_1 is the total energy of starting configuration and E_2 is the
    total energy of the end configuration. Q_s is a constant that depends
    on the species. If an isolated solute is diffusing in a host material
    E_1 = E_2. Thus, Q_s represents the barrier of an isolated solute.

    :param dilute_barrier: Dictionary representing the dilute barrier for
        each species. Example {'Al': 0.05, 'Mg': 0.03}
    """

    def __init__(self, dilute_barrier: Dict[str, float]):
        super().__init__()
        self.dilute_barrier = dilute_barrier

    def __call__(self, atoms: Atoms, system_changes: Sequence[SystemChange]) -> float:
        E1 = atoms.calc.updater.get_energy()
        E2 = atoms.calc.get_energy_given_change(system_changes)

        # Extract the jumping species. The system change should involve
        # a vacancy and another element. Thus, the jumping species is the
        # symbol in system_change that is not a vacancy
        old_symb, new_symb = system_changes[0].old_symb, system_changes[0].new_symb

        # Confirm that one of the symbols is a vacancy
        assert old_symb == 'X' or new_symb == 'X'

        jumping_symb = old_symb if old_symb != 'X' else new_symb
        Ea = self.dilute_barrier[jumping_symb] + 0.5 * (E2 - E1)
        return Ea



# pylint: disable=undefined-variable
import logging
from deprecated import deprecated
from .template_filters import *
from .atoms_manager import *
from .convexhull import *
from .data_manager import *
from .structure_mapper import *
from .evaluate import *
from .new_struct import NewStructures  # Promotion for convenience
from .svd import *
from .concentration import *
from .regression_old import *
from .cluster_coverage import *

from . import cluster
from . import new_struct
from . import corr_func
from . import settings
from . import basis_function
from . import montecarlo
from . import mp_logger
from . import tools
from . import template_atoms
from . import concentration

__version__ = '0.10.0'
logger = logging.getLogger(__name__)


@deprecated(version='0.10.0', reason='import CEBulk from clease.settings instead')
def CEBulk(*args, **kwargs):
    return settings.CEBulk(*args, **kwargs)


@deprecated(version='0.10.0', reason='import CECrystal from clease.settings instead')
def CECrystal(*args, **kwargs):
    return settings.CECrystal(*args, **kwargs)


@deprecated(version='0.10.0', reason='import CESlab from clease.settings instead')
def CESlab(*args, **kwargs):
    return settings.CESlab(*args, **kwargs)


ADDITIONAL = ('settings', 'basis_function', 'corr_func', 'new_struct', 'NewStructures', 'cluster',
              'tools', 'montecarlo', 'mp_logger', 'template_atoms', 'CEBulk', 'CECrystal', 'CESlab',
              'concentration')

__all__ = (template_filters.__all__ + atoms_manager.__all__ + concentration.__all__ +
           evaluate.__all__ + convexhull.__all__ + data_manager.__all__ + structure_mapper.__all__ +
           svd.__all__ + regression_old.__all__ + cluster_coverage.__all__ + ADDITIONAL)



from clease.montecarlo.observers import MCObserver
from clease.tools import add_file_extension
from ase.io.trajectory import TrajectoryWriter


class Snapshot(MCObserver):
    """Store a snapshot in a trajectory file.

    Parameters:

    fname: str
        Name of the trajectory file. Adds extension '.traj' if none is given.

    atoms: Atoms
        Instance of the atoms objected modofied by the MC object
    """

    name = "Snapshot"

    def __init__(self, fname="snapshot.traj", atoms=None):
        super().__init__()
        full_fname = add_file_extension(fname, '.traj')
        if atoms is None:
            raise ValueError("No atoms object given!")
        self.atoms = atoms
        self.traj = TrajectoryWriter(full_fname, mode="w")
        self.fname = full_fname

    def __call__(self, system_changes):
        """Write a snapshot to a .traj file.

        Parameters:

        system_changes: list
            System changes. See doc-string of
            `clease.montecarlo.observers.MCObserver`
        """
        self.traj.write(self.atoms)

    def reset(self):
        self.traj = TrajectoryWriter(self.fname, mode="w")
